# Copyright 2020-2021 The MathWorks, Inc.

import json
import os
import sys
from json.decoder import JSONDecodeError

from matlab_proxy.util.mwi import environment_variables as mwi_env

from . import logger as mwi_logger

logger = mwi_logger.get()


def get():
    """Returns a dict containing custom http headers to inject into responses from the
    MATLAB Embedded Connector and to the static files in 'gui' folder.

    Parses valid JSON data if provided as a string or in a JSON file.

    Returns:
        Dict: Representing custom headers.
    """
    env_var = __get_custom_header_env_var()
    # custom_headers_path can contain valid JSON data as a string or a path to a file containing valid JSON
    custom_headers_path = os.getenv(env_var)

    # If the environment variable is not present
    if custom_headers_path is None:
        logger.debug(
            f"Environment variable {env_var} is not set, hence no custom HTTP headers are applied."
        )
        return dict()

    logger.info(
        f"Envinronment variable {env_var} is set and has the value: {custom_headers_path}"
    )
    is_file = os.path.isfile(custom_headers_path)

    if is_file and __check_file_validity(custom_headers_path):
        return __get_file_contents(custom_headers_path)

    else:
        exception_statement = __get_exception_statement()
        try:
            # Valid JSON data as a string
            data = json.loads(custom_headers_path)
            logger.info(
                f"Successfully parsed HTTP headers from the environment variable {env_var}\nThe parsed HTTP headers are:\n{data}\n"
            )
            return data

        except JSONDecodeError as json_decode_error:
            # Invalid JSON data as a string.
            print(
                exception_statement.format(
                    env_var=__get_custom_header_env_var(),
                    env_var_value=custom_headers_path,
                )
            )
            logger.error(
                f"Failed with {type(json_decode_error).__name__} exception. Check the value in {env_var} environment variable for syntax errors."
            )
            sys.exit(1)

        except Exception as e:
            print(
                exception_statement.format(
                    env_var=__get_custom_header_env_var(),
                    env_var_value=custom_headers_path,
                )
            )
            logger.error(
                f"Raised {type(e).__name__} exception when parsing JSON data in the environment variable {env_var}"
            )
            sys.exit(1)


def __check_file_validity(custom_headers_path):
    """Checks the file at custom_headers_path for the read access to the file for the current python process.

    Args:
        custom_headers_path (String): File path of custom headers.

    Raises:
        OSError: When the current python process doesn't have read access to the custom_headers_path
    """
    exception_statement = __get_exception_statement()
    logger.debug(
        f"The value in environment variable {__get_custom_header_env_var()} contains a path to a file."
    )
    try:
        if not os.access(custom_headers_path, os.R_OK):
            raise OSError

        logger.debug(
            f"Current python process has read access rights to the file at: {custom_headers_path}"
        )
        return True

    except OSError as os_error:
        print(
            f"\n{type(os_error).__name__}: Permission denied. Cannot read file: {custom_headers_path}\n"
            + exception_statement.format(
                env_var=__get_custom_header_env_var(),
                env_var_value=custom_headers_path,
            )
        )
        logger.error(
            f"{type(os_error).__name__}: Permission denied. For the current python process, check read access rights for the file: {custom_headers_path}."
        )
        sys.exit(1)

    except Exception as e:
        print(
            f"\n{type(e).__name__} exception raised with error message: {e.args}\n"
            + exception_statement.format(
                env_var=__get_custom_header_env_var(),
                env_var_value=custom_headers_path,
            )
        )
        logger.error(
            f"{type(e).__name__} exception raised when checking read access rights to the file at: {custom_headers_path}"
        )
        sys.exit(1)


def __get_file_contents(custom_headers_path):
    """Reads the file containing custom headers and returns a dict.
    Raises JSONDecodeError if the JSON data in the file is not valid.

    Args:
        custom_headers_path (String): File path of custom headers.

    Raises:
        JSONDecodeError: when the contents of the file containing custom headers
        does not have valid JSON data.

    Returns:
        Dict: Containing custom headers.
    """

    custom_headers = None
    exception_statement = __get_exception_statement()
    try:
        with open(custom_headers_path, "r") as json_file:
            custom_headers = json.load(json_file)
            logger.info(
                f"Successfully parsed HTTP headers present in the file: {custom_headers_path}"
            )
            logger.debug(f"The parsed HTTP headers are :\n{custom_headers}\n")
            return custom_headers

    except JSONDecodeError as json_decode_error:
        print(
            f"\n{type(json_decode_error).__name__} exception raised with error message: {json_decode_error.args[0]}\n"
            + exception_statement.format(
                env_var=__get_custom_header_env_var(),
                env_var_value=custom_headers_path,
            )
        )
        logger.error(
            f"{type(json_decode_error).__name__}: Failed to parse JSON data in the file: {custom_headers_path}.\nCheck contents of the file for any syntax errors."
        )
        sys.exit(1)

    except Exception as e:
        print(
            f"\nFailed with {type(e).__name__} exception with error message:\n{e.args}"
            + exception_statement.format(
                env_var=__get_custom_header_env_var(),
                env_var_value=custom_headers_path,
            )
        )
        logger.error(
            f"Failed with {type(e).__name__}  when reading JSON data from the file: {custom_headers_path}"
        )
        sys.exit(1)


def __get_custom_header_env_var():
    """Returns the name of the environment variable which contains path
    to the file containing custom headers.

    Returns:
        String: Environment variable containing path to the file containing custom headers.
    """
    return mwi_env.get_env_name_custom_http_headers()


def __get_exception_statement():
    """Returns a generic exception statement pertaining to HTTP headers.

    Returns:
        String: Containing a generic exception statement to print to stdout.
    """
    exception_statement = """HTTP headers defined in the environment variable {env_var} are not valid.\n
    The value in {env_var} is: {env_var_value}\n
    Please provide either valid JSON data as a string or a path to a file with read access containing valid JSON data."""

    return exception_statement



from matlab_proxy.util import mwi


def test_get_data_to_eval_mcode():
    necessary_keys = ["uuid", "messages", "computeToken"]
    mcode = "exit"
    data = mwi.embedded_connector.helpers.get_data_to_eval_mcode(mcode)

    assert set(necessary_keys).issubset(set(data.keys()))
    assert data["messages"]["Eval"][0]["mcode"] == mcode


def test_get_data_to_feval_mcode():
    necessary_keys = ["uuid", "messages", "computeToken"]
    m_function = "round"
    args = [3.135, 2]
    nargout = 2
    data = mwi.embedded_connector.helpers.get_data_to_feval_mcode(
        m_function, *args, nargout=nargout
    )

    assert set(necessary_keys).issubset(set(data.keys()))
    feval_data = data["messages"]["FEval"][0]

    assert feval_data["function"] == m_function
    assert feval_data["arguments"] == args
    assert feval_data["nargout"] == nargout



# Copyright 2020-2021 The MathWorks, Inc.

import os
import shutil
import socket
import ssl
import sys
import tempfile
import uuid
import xml.etree.ElementTree as ET
from pathlib import Path

import matlab_proxy
from matlab_proxy.util import mwi
from matlab_proxy.util.mwi import environment_variables as mwi_env

logger = mwi.logger.get()


def get_matlab_path():
    which_matlab = shutil.which("matlab")
    if which_matlab is None:
        return None
    return Path(which_matlab).resolve().parent.parent


def get_matlab_version(matlab_path):
    """Get the MATLAB Release version in this image"""

    if matlab_path is None:
        return None

    tree = ET.parse(matlab_path / "VersionInfo.xml")
    root = tree.getroot()
    return root.find("release").text


def get_ws_env_settings():
    ws_env = (os.getenv("WS_ENV") or "").lower()
    ws_env_suffix = f"-{ws_env}" if "integ" in ws_env else ""

    return ws_env, ws_env_suffix


def get_dev_settings(config):
    devel_file = Path(__file__).resolve().parent / "./devel.py"
    mwi_config_folder = get_test_temp_dir()
    ws_env, ws_env_suffix = get_ws_env_settings()
    return {
        "matlab_path": Path(),
        "matlab_version": "R2020b",
        "matlab_cmd": [
            "python",
            "-u",
            str(devel_file),
            "matlab",
        ],
        "create_xvfb_cmd": create_xvfb_cmd,
        "base_url": os.environ.get(mwi_env.get_env_name_base_url(), ""),
        "app_port": os.environ.get(mwi_env.get_env_name_app_port(), 8000),
        "host_interface": os.environ.get(mwi_env.get_env_name_app_host(), "127.0.0.1"),
        "mwapikey": str(uuid.uuid4()),
        "matlab_protocol": "http",
        "matlab_display": ":1",
        "nlm_conn_str": os.environ.get(mwi_env.get_env_name_network_license_manager()),
        "matlab_config_file": mwi_config_folder / "proxy_app_config.json",
        "ws_env": ws_env,
        "mwa_api_endpoint": f"https://login{ws_env_suffix}.mathworks.com/authenticationws/service/v4",
        "mhlm_api_endpoint": f"https://licensing{ws_env_suffix}.mathworks.com/mls/service/v1/entitlement/list",
        "mwa_login": f"https://login{ws_env_suffix}.mathworks.com",
        "mwi_custom_http_headers": mwi.custom_http_headers.get(),
        "env_config": mwi.validators.validate_env_config(config),
        "ssl_context": None,
        "mwi_logs_root_dir": mwi_config_folder / "ports",
        "mwi_proxy_lock_file_name": "mwi_proxy.lock",
        "mw_context_tags": get_mw_context_tags(matlab_proxy.get_default_config_name()),
        "mwi_server_url": None,
    }


def get(config_name=matlab_proxy.get_default_config_name(), dev=False):
    """Returns the settings specific to the environment in which the server is running in
    If the environment variable 'TEST' is set  to true, will make some changes to the dev settings.

    Args:
        config : Dictionary as specified by the default_configuration.py file. Used to customize the app.
        dev (bool, optional): development environment. Defaults to False.

    Returns:
        Dict: Containing data on how to start MATLAB among other information.
    """

    if dev:
        settings = get_dev_settings(config_name)

        # If running tests using Pytest, it will set environment variable TEST to true before running tests.
        # Will make test env specific changes before returning the settings.
        if mwi_env.is_testing_mode_enabled():

            # Set ready_delay value to 0 for faster fake MATLAB startup.
            ready_delay = ["--ready-delay", "0"]
            matlab_cmd = settings["matlab_cmd"]
            matlab_cmd[4:4] = ready_delay
            settings["matlab_cmd"] = matlab_cmd

            # Randomly picks an available port and updates the value of settings['app_port'] .
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.bind(("", 0))
            settings["app_port"] = s.getsockname()[1]
            s.close()

            # Set NLM Connection string. Server will start using this connection string for licensing
            settings["nlm_conn_str"] = "abc@nlm"

        return settings

    else:
        matlab_startup_file = str(
            Path(__file__).resolve().parent / "matlab" / "startup.m"
        )
        matlab_path = get_matlab_path()
        ws_env, ws_env_suffix = get_ws_env_settings()

        ssl_key_file, ssl_cert_file = mwi.validators.validate_ssl_key_and_cert_file(
            os.getenv(mwi_env.get_env_name_ssl_key_file(), None),
            os.getenv(mwi_env.get_env_name_ssl_cert_file(), None),
        )

        # All config related to matlab-proxy will be saved to user's home folder.
        # This will allow for other user's to launch the integration from the same system
        # and not have their config's overwritten.
        mwi_config_folder = Path.home() / ".matlab" / "MWI"
        return {
            "matlab_path": matlab_path,
            "matlab_version": get_matlab_version(matlab_path),
            "matlab_cmd": [
                "matlab",
                "-nosplash",
                "-nodesktop",
                "-softwareopengl",
                "-r",
                f"try; run('{matlab_startup_file}'); catch; end;",
            ],
            "create_xvfb_cmd": create_xvfb_cmd,
            "base_url": mwi.validators.validate_base_url(
                os.getenv(mwi_env.get_env_name_base_url(), "")
            ),
            "app_port": mwi.validators.validate_app_port_is_free(
                os.getenv(mwi_env.get_env_name_app_port())
            ),
            "host_interface": os.environ.get(mwi_env.get_env_name_app_host()),
            "mwapikey": str(uuid.uuid4()),
            "matlab_protocol": "https",
            "nlm_conn_str": mwi.validators.validate_mlm_license_file(
                os.environ.get(mwi_env.get_env_name_network_license_manager())
            ),
            "matlab_config_file": mwi_config_folder / "proxy_app_config.json",
            "ws_env": ws_env,
            "mwa_api_endpoint": f"https://login{ws_env_suffix}.mathworks.com/authenticationws/service/v4",
            "mhlm_api_endpoint": f"https://licensing{ws_env_suffix}.mathworks.com/mls/service/v1/entitlement/list",
            "mwa_login": f"https://login{ws_env_suffix}.mathworks.com",
            "mwi_custom_http_headers": mwi.custom_http_headers.get(),
            "env_config": mwi.validators.validate_env_config(config_name),
            "ssl_context": get_ssl_context(
                ssl_cert_file=ssl_cert_file, ssl_key_file=ssl_key_file
            ),
            # This directory will be used to store connector.securePort(matlab_ready_file) and its corresponding files. This will be
            # a central place to store logs of all the running instances of MATLAB launched by matlab-proxy
            "mwi_logs_root_dir": mwi_config_folder / "ports",
            # Name of the lock file which will be created by this instance of matlab-proxy process.
            "mwi_proxy_lock_file_name": "mwi_proxy.lock",
            "mw_context_tags": get_mw_context_tags(config_name),
            # The url where the matlab-proxy server is accessible at
            "mwi_server_url": None,
        }


def get_mw_context_tags(extension_name):
    """Returns a string which combines existing MW_CONTEXT_TAGS value and context tags
    specific to where matlab-proxy is being launched from.

    Returns:
        str: Which combines existing MW_CONTEXT_TAGS with one from matlab-proxy.
    """
    existing_mw_context_tags = os.getenv("MW_CONTEXT_TAGS", "")

    if existing_mw_context_tags:
        logger.debug(f'Existing MW_CONTEXT_TAGS:"{existing_mw_context_tags}"')
        existing_mw_context_tags += ","

    mwi_context_tags = matlab_proxy.get_mwi_ddux_value(extension_name)
    logger.debug(f'DDUX value for matlab-proxy "{mwi_context_tags}"')

    combined_context_tags = existing_mw_context_tags + mwi_context_tags
    logger.debug(
        f'Combined DDUX value to be used for MATLAB process: "{combined_context_tags}"'
    )

    return combined_context_tags


def create_xvfb_cmd():
    """Creates the Xvfb command with a write descriptor.

    Returns:
        List: Containing 2 lists.

    The second List contains a read and a write descriptor.
    The first List is the command to launch Xvfb process with the same write descriptor(from the first list) embedded in the command.
    """
    # Using os.pipe() can lead to race conditions (ie.usage of same set of file descriptors between 2 processes)
    # when called in quick succession and also when running tests.
    # Using os.pipe2() with the flag os.O_NONBLOCK will avoid race conditions.
    dpipe = os.pipe2(os.O_NONBLOCK)

    # Allow child process to use the file descriptor created by parent.
    os.set_inheritable(dpipe[1], True)

    xvfb_cmd = [
        "Xvfb",
        "-displayfd",
        # Write descriptor
        str(dpipe[1]),
        "-screen",
        "0",
        "1600x1200x24",
        "-dpi",
        "100",
        # "-ac",
        "-extension",
        "RANDR",
        # "+render",
        # "-noreset",
    ]

    return xvfb_cmd, dpipe


def get_test_temp_dir():
    """The temp directory to be used by tests"""
    test_temp_dir = Path(tempfile.gettempdir()) / "MWI" / "tests"
    test_temp_dir.mkdir(parents=True, exist_ok=True)
    return test_temp_dir


def get_ssl_context(ssl_cert_file, ssl_key_file):
    """Creates an SSL CONTEXT for use with the TCP Site"""

    # The certfile string must be the path to a single file in PEM format containing the
    # certificate as well as any number of CA certificates needed to establish the certificate’s authenticity.
    # The keyfile string, if present, must point to a file containing the private key in.
    # Otherwise the private key will be taken from certfile as well.
    import traceback

    if ssl_cert_file != None:
        try:
            ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            ssl_context.load_cert_chain(ssl_cert_file, ssl_key_file)
            logger.debug(f"Using SSL certification!")
        except Exception as e:
            # Something was wrong with the certificates provided
            logger.error("SSL certificates provided are invalid. Aborting...")
            traceback.print_exc()
            logger.info("==== Fatal error : ===")
            print(e)
            # printing stack trace
            logger.info("======================")
            sys.exit(1)
    else:
        ssl_context = None

    return ssl_context



from matlab_proxy import util


def test_is_python_version_newer_than_3_6():
    """Test to check python version is 3.7 or later"""
    assert util.is_python_version_newer_than_3_6() is not None


def test_get_supported_termination_signals():
    assert len(util.__get_supported_termination_signals()) >= 1


def test_add_signal_handlers(loop):
    """Test to check if signal handlers are being added to asyncio loop

    Args:
        loop (asyncio loop): In built-in pytest fixture.
    """

    loop = util.add_signal_handlers(loop)
    assert loop._signal_handlers is not None
    assert loop._signal_handlers.items() is not None



# Copyright 2020-2021 The MathWorks, Inc.
"""Functions to access & control the logging behavior of the app
"""

import logging
import os

from . import environment_variables as mwi_env


def get(init=False):
    """Get the logger used by this application.
        Set init=True to initialize the logger
    Returns:
        Logger: The logger used by this application.
    """
    if init is True:
        return __set_logging_configuration()

    return __get_mw_logger()


def __get_mw_logger_name():
    """Name of logger used by the app

    Returns:
        String: The name of the Logger.
    """
    return "MATLABProxyApp"


def __get_mw_logger():
    """Returns logger for use in this app.

    Returns:
        Logger: A logger object
    """
    return logging.getLogger(__get_mw_logger_name())


def __set_logging_configuration():
    """Sets the logging environment for the app

    Returns:
        Logger: Logger object with the set configuration.
    """
    # query for user specified environment variables
    log_level, log_file = __query_environment()

    ## Set logging object
    logger = __get_mw_logger()
    if log_file is not None:
        logger.info(f"Initializing logger with log_file:{log_file}")
        file_handler = logging.FileHandler(filename=log_file)
        formatter = logging.Formatter(
            fmt="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        file_handler.setFormatter(formatter)
        file_handler.setLevel(log_level)
        logger.addHandler(file_handler)

    # log_level is either set by environment or is the default value.
    logger.info(f"Initializing logger with log_level: {log_level}")
    logger.setLevel(log_level)

    # Allow other libraries used by this integration to
    # also print their logs at the specified level
    logging.basicConfig(level=log_level)

    return logger


def get_environment_variable_names():
    """Helper to return names of environment variables queried.

    Returns:
        tuple: name of environment variable to control log level,
                name of environment variable to control logging to file
    """
    __log_file_environment_variable_name = mwi_env.get_env_name_log_file()
    __log_level_environment_variable_name = mwi_env.get_env_name_logging_level()
    return __log_level_environment_variable_name, __log_file_environment_variable_name


def __get_default_log_level():
    """The default logging level used by this application.

    Returns:
        String: The default logging level
    """
    return "INFO"


def __query_environment():
    """Private function to query environment variables
        to control logging

    Returns:
        tuple: Log level & Log file as found in the environment
    """
    env_var_log_level, env_var_log_file = get_environment_variable_names()
    log_level = os.environ.get(env_var_log_level, __get_default_log_level())
    log_file = os.environ.get(env_var_log_file)
    return log_level, log_file



# Copyright 2021 The MathWorks, Inc.
"""This file tests methods present in matlab_proxy/util/mwi_logger.py
"""

import logging
import os

from matlab_proxy.util.mwi import logger as mwi_logger


def test_get():
    """This test checks if the get method returns a logger with expected name"""
    logger = mwi_logger.get()
    # Okay to use hidden API for testing only.
    assert logger.name == mwi_logger.__get_mw_logger_name()


def test_get_mw_logger_name():
    """Test to lock down the name of the logger used."""
    # Okay to use hidden API for testing only.
    assert "MATLABProxyApp" == mwi_logger.__get_mw_logger_name()


def test_get_with_no_environment_variables(monkeypatch):
    """This test checks if the get method returns a logger with default settings"""
    # Delete the environment variables if they do exist
    env_names_list = mwi_logger.get_environment_variable_names()
    monkeypatch.delenv(env_names_list[0], raising=False)
    monkeypatch.delenv(env_names_list[1], raising=False)

    logger = mwi_logger.get(init=True)
    assert logger.isEnabledFor(logging.INFO) == True
    assert len(logger.handlers) == 0


def test_get_with_environment_variables(monkeypatch, tmp_path):
    """This test checks if the get method returns a logger with default settings"""
    env_names_list = mwi_logger.get_environment_variable_names()
    monkeypatch.setenv(env_names_list[0], "CRITICAL")
    monkeypatch.setenv(env_names_list[1], str(tmp_path / "testing123.log"))

    logger = mwi_logger.get(init=True)

    # Verify that environment variable controlling level is respected
    assert logger.isEnabledFor(logging.CRITICAL) == True

    # Verify that environment variable setting the file is respected
    assert len(logger.handlers) == 1
    assert os.path.basename(logger.handlers[0].baseFilename) == "testing123.log"



# Copyright 2020-2021 The MathWorks, Inc.

import asyncio
import errno
import json
import logging
import os
import pty
import socket
from collections import deque
from datetime import datetime, timedelta, timezone

from matlab_proxy import util
from matlab_proxy.util import mw, mwi
from matlab_proxy.util.mwi import environment_variables as mwi_env
from matlab_proxy.util.mwi.exceptions import (
    EmbeddedConnectorError,
    EntitlementError,
    InternalError,
    LicensingError,
    MatlabInstallError,
    OnlineLicensingError,
    log_error,
)

logger = mwi.logger.get()


class AppState:
    """A Class which represents the state of the App.
    This class handles state of MATLAB, MATLAB Licensing and Xvfb.
    """

    def __init__(self, settings):
        """Parameterized constructor for the AppState class.
        Initializes member variables and checks for an existing MATLAB installation.

        Args:
            settings (Dict): Represents the settings required for managing MATLAB, Licensing and Xvfb.
        """
        self.settings = settings
        self.processes = {"matlab": None, "xvfb": None}

        # The port on which MATLAB(launched by this matlab-proxy process) starts on.
        self.matlab_port = None

        # The file created and written by MATLAB's Embedded connector to signal readiness.
        self.matlab_ready_file = None

        # The directory in which the instance of MATLAB (launched by this matlab-proxy process) will write logs to.
        self.matlab_ready_file_dir = None

        # The file created by this instance of matlab-proxy to signal to other matlab-proxy processes
        # that this self.matlab_port will be used by this instance.
        self.mwi_proxy_lock_file = None
        self.licensing = None
        self.tasks = {}
        self.logs = {
            "matlab": deque(maxlen=200),
        }
        self.error = None
        # Start in an error state if MATLAB is not present
        if not self.is_matlab_present():
            self.error = MatlabInstallError("'matlab' executable not found in PATH")
            logger.error("'matlab' executable not found in PATH")
            return

    def __get_cached_licensing_file(self):
        """Get the cached licensing file

        Returns:
            Path : Path object to cached licensing file
        """
        return self.settings["matlab_config_file"]

    def __delete_cached_licensing_file(self):
        """Deletes the cached licensing file"""
        try:
            logger.info(f"Deleting any cached licensing files!")
            os.remove(self.__get_cached_licensing_file())
        except FileNotFoundError:
            # The file being absent is acceptable.
            pass

    def __reset_and_delete_cached_licensing(self):
        """Reset licensing variable of the class and removes the cached licensing file."""
        logger.info(f"Resetting cached licensing information...")
        self.licensing = None
        self.__delete_cached_licensing_file()

    async def __update_and_persist_licensing(self):
        """Update entitlements from mhlm servers and persist licensing

        Returns:
            Boolean: True when entitlements were updated and persisted successfully. False otherwise.
        """
        successful_update = await self.update_entitlements()
        if successful_update:
            self.persist_licensing()
        else:
            self.__reset_and_delete_cached_licensing()
        return successful_update

    async def init_licensing(self):
        """Initialize licensing from environment variable or cached file.

        Greater precedence is given to value specified in environment variable MLM_LICENSE_FILE
            If specified, this function will delete previously cached licensing information.
            This enforces a clear understanding of what was used to initialize licensing.
            The contents of the environment variable are NEVER cached.
        """

        # Default value
        self.licensing = None

        # NLM Connection String set in environment
        if self.settings["nlm_conn_str"] is not None:
            nlm_licensing_str = self.settings["nlm_conn_str"]
            logger.debug(f"Found NLM:[{nlm_licensing_str}] set in environment")
            logger.debug(f"Using NLM string to connect ... ")
            self.licensing = {
                "type": "nlm",
                "conn_str": nlm_licensing_str,
            }
            self.__delete_cached_licensing_file()

        # If NLM connection string is not present, then look for persistent LNU info
        elif self.__get_cached_licensing_file().exists():
            with open(self.__get_cached_licensing_file(), "r") as f:
                logger.debug("Found cached licensing information...")
                try:
                    # Load can throw if the file is empty for some reason.
                    licensing = json.loads(f.read())
                    if licensing["type"] == "nlm":
                        # Note: Only NLM settings entered in browser were cached.
                        self.licensing = {
                            "type": "nlm",
                            "conn_str": licensing["conn_str"],
                        }
                    elif licensing["type"] == "mhlm":
                        self.licensing = {
                            "type": "mhlm",
                            "identity_token": licensing["identity_token"],
                            "source_id": licensing["source_id"],
                            "expiry": licensing["expiry"],
                            "email_addr": licensing["email_addr"],
                            "first_name": licensing["first_name"],
                            "last_name": licensing["last_name"],
                            "display_name": licensing["display_name"],
                            "user_id": licensing["user_id"],
                            "profile_id": licensing["profile_id"],
                            "entitlements": [],
                            "entitlement_id": licensing.get("entitlement_id"),
                        }

                        expiry_window = datetime.strptime(
                            self.licensing["expiry"], "%Y-%m-%dT%H:%M:%S.%f%z"
                        ) - timedelta(hours=1)

                        if expiry_window > datetime.now(timezone.utc):
                            successful_update = (
                                await self.__update_and_persist_licensing()
                            )
                            if successful_update:
                                logger.debug("Successful re-use of cached information.")
                        else:
                            self.__reset_and_delete_cached_licensing()
                    else:
                        # Somethings wrong, licensing is neither NLM or MHLM
                        self.__reset_and_delete_cached_licensing()
                except Exception as e:
                    self.__reset_and_delete_cached_licensing()

    async def get_matlab_state(self):
        """Determine the state of MATLAB to be down/starting/up.

        Returns:
            String: Status of MATLAB. Returns either up, down or starting.
        """

        matlab = self.processes["matlab"]
        xvfb = self.processes["xvfb"]

        # MATLAB process never started
        if matlab is None:
            return "down"
        # MATLAB process previously started, but not currently running
        elif matlab.returncode is not None:
            return "down"
        # Xvfb never started
        elif xvfb is None:
            return "down"
        # Xvfb process previously started, but not running
        elif xvfb.returncode is not None:
            return "down"
        else:
            # If execution reaches this else block, it implies that:
            # 1) MATLAB process has started.
            # 2) Embedded connector has not started yet.

            # So, even if the embedded connector's status is 'down', we'll
            # return as 'starting' because the MATLAB process itself has been created
            # and matlab-proxy is waiting for the embedded connector to start serving content.
            status = await mwi.embedded_connector.request.get_state(
                self.settings["mwi_server_url"]
            )
            if status == "down":
                status = "starting"

            return status

    async def set_licensing_nlm(self, conn_str):
        """Set the licensing type to NLM and the connection string."""

        # TODO Validate connection string
        self.licensing = {"type": "nlm", "conn_str": conn_str}
        self.persist_licensing()

    async def set_licensing_mhlm(
        self,
        identity_token,
        email_addr,
        source_id,
        entitlements=[],
        entitlement_id=None,
    ):
        """Set the licensing type to MHLM and the details.

        Args:
            identity_token (String): Identity token of the user.
            email_addr (String): Email address of the user.
            source_id (String): Unique random string generated for the user.
            entitlements (list, optional): Eligible Entitlements of the user. Defaults to [].
            entitlement_id (String, optional): ID of an entitlement. Defaults to None.
        """

        try:

            token_data = await mw.fetch_expand_token(
                self.settings["mwa_api_endpoint"], identity_token, source_id
            )

            self.licensing = {
                "type": "mhlm",
                "identity_token": identity_token,
                "source_id": source_id,
                "expiry": token_data["expiry"],
                "email_addr": email_addr,
                "first_name": token_data["first_name"],
                "last_name": token_data["last_name"],
                "display_name": token_data["display_name"],
                "user_id": token_data["user_id"],
                "profile_id": token_data["profile_id"],
                "entitlements": entitlements,
                "entitlement_id": entitlement_id,
            }

            successful_update = await self.__update_and_persist_licensing()
            if successful_update:
                logger.debug("Login successful, persisting login information.")

        except OnlineLicensingError as e:
            self.error = e
            self.licensing = {
                "type": "mhlm",
                "email_addr": email_addr,
            }
            log_error(logger, e)

    def unset_licensing(self):
        """Unset the licensing."""

        self.licensing = None

        # If the error was due to licensing, clear it
        if isinstance(self.error, LicensingError):
            self.error = None

    def is_licensed(self):
        """Is MATLAB licensing configured?

        Returns:
            Boolean: True if MATLAB is Licensed. False otherwise.
        """

        if self.licensing is not None:
            if self.licensing["type"] == "nlm":
                if self.licensing["conn_str"] is not None:
                    return True
            elif self.licensing["type"] == "mhlm":
                if (
                    self.licensing.get("identity_token") is not None
                    and self.licensing.get("source_id") is not None
                    and self.licensing.get("expiry") is not None
                    and self.licensing.get("entitlement_id") is not None
                ):
                    return True
        return False

    def is_matlab_present(self):
        """Is MATLAB install accessible?

        Returns:
            Boolean: True if MATLAB is present in the system. False otherwise.
        """

        return self.settings["matlab_path"] is not None

    async def update_entitlements(self):
        """Speaks to MW and updates MHLM entitlements

        Raises:
            InternalError: When licensing is None or when licensing type is not MHLM.

        Returns:
            Boolean: True if update was successful
        """
        if self.licensing is None or self.licensing["type"] != "mhlm":
            raise InternalError(
                "MHLM licensing must be configured to update entitlements!"
            )

        try:
            # Fetch an access token
            access_token_data = await mw.fetch_access_token(
                self.settings["mwa_api_endpoint"],
                self.licensing["identity_token"],
                self.licensing["source_id"],
            )

            # Fetch entitlements
            entitlements = await mw.fetch_entitlements(
                self.settings["mhlm_api_endpoint"],
                access_token_data["token"],
                self.settings["matlab_version"],
            )

        except OnlineLicensingError as e:
            self.error = e
            log_error(logger, e)
            return False
        except EntitlementError as e:
            self.error = e
            log_error(logger, e)
            self.licensing["identity_token"] = None
            self.licensing["source_id"] = None
            self.licensing["expiry"] = None
            self.licensing["first_name"] = None
            self.licensing["last_name"] = None
            self.licensing["display_name"] = None
            self.licensing["user_id"] = None
            self.licensing["profile_id"] = None
            self.licensing["entitlements"] = []
            self.licensing["entitlement_id"] = None
            return False

        self.licensing["entitlements"] = entitlements

        # If there is only one non-expired entitlement, set it as active
        # TODO Also, for now, set the first entitlement as active if there are multiple
        self.licensing["entitlement_id"] = entitlements[0]["id"]

        # Successful update
        return True

    def persist_licensing(self):
        """Saves licensing information to file"""
        if self.licensing is None:
            self.__delete_cached_licensing_file()

        elif self.licensing["type"] in ["mhlm", "nlm"]:
            logger.debug("Saving licensing information...")
            cached_licensing_file = self.__get_cached_licensing_file()
            cached_licensing_file.parent.mkdir(parents=True, exist_ok=True)
            with open(cached_licensing_file, "w") as f:
                f.write(json.dumps(self.licensing))

    def prepare_lock_files_for_MATLAB_launch(self):
        """Finds and reserves a free port for MATLAB Embedded Connector in the allowed range.
        Creates the lock file to prevent any other matlab-proxy process to use the reserved port of this
        process.

        Raises:
            e: socket.error if the exception raised is other than port already occupied.
        """

        # NOTE It is not guranteed that the port will remain free!
        # FIXME Because of https://github.com/http-party/node-http-proxy/issues/1342 the
        # node application in development mode always uses port 31515 to bypass the
        # reverse proxy. Once this is addressed, remove this special case.
        if (
            mwi_env.is_development_mode_enabled()
            and not mwi_env.is_testing_mode_enabled()
        ):
            return 31515
        else:

            # TODO If MATLAB Connector is enhanced to allow any port, then the
            # following can be used to get an unused port instead of the for loop and
            # try-except.
            # s.bind(("", 0))
            # self.matlab_port = s.getsockname()[1]

            for port in mw.range_matlab_connector_ports():
                try:
                    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                    s.bind(("", port))

                    # Use the app_port number to identify the server as that is user visible
                    matlab_ready_file_dir = self.settings["mwi_logs_root_dir"] / str(
                        self.settings["app_port"]
                    )

                    mwi_proxy_lock_file = (
                        matlab_ready_file_dir
                        / self.settings["mwi_proxy_lock_file_name"]
                    )

                    matlab_ready_file = matlab_ready_file_dir / "connector.securePort"

                    # Check if the mwi_proxy_lock_file exists.
                    # Implies there was a competing matlab-proxy process which found the same port before this process
                    if mwi_proxy_lock_file.exists():
                        logger.debug(
                            f"Skipping port number {port} for MATLAB as lock file already exists at {mwi_proxy_lock_file}"
                        )
                        s.close()

                    else:
                        # Create a folder to hold the matlab_ready_file that will be created by MATLAB to signal readiness.
                        # This is the same folder to which MATLAB will write logs to.
                        matlab_ready_file_dir.mkdir(parents=True, exist_ok=True)

                        # Creating the mwi_proxy.lock file to indicate to any other matlab-proxy processes
                        # that this self.matlab_port number is taken up by this process.
                        with open(mwi_proxy_lock_file, "w") as lock_file:
                            lock_file.write(self.settings["mwi_server_url"] + "/")

                        # Update member variables of AppState class

                        # Store the port number on which MATLAB will be launched for this matlab-proxy process.
                        self.matlab_port = port
                        self.mwi_proxy_lock_file = mwi_proxy_lock_file
                        self.matlab_ready_file_dir = matlab_ready_file_dir
                        self.matlab_ready_file = matlab_ready_file
                        s.close()

                        logger.debug(
                            f"MATLAB_LOG_DIR:{str( self.matlab_ready_file_dir )}"
                        )
                        logger.debug(f"MATLAB_READY_FILE:{str(self.matlab_ready_file)}")
                        logger.debug(
                            f"Created MWI proxy lock file for this matlab-proxy process at {self.mwi_proxy_lock_file}"
                        )
                        return

                except socket.error as e:
                    if e.errno != errno.EADDRINUSE:
                        raise e

    async def __setup_env_for_matlab(self) -> dict:
        """Configure the environment variables required for launching MATLAB by matlab-proxy.

        Returns:
            [dict]: Containing keys as the Env variable names and values are its corresponding values.
        """
        matlab_env = os.environ.copy()
        # Env setup related to licensing
        if self.licensing["type"] == "mhlm":
            try:
                # Request an access token
                access_token_data = await mw.fetch_access_token(
                    self.settings["mwa_api_endpoint"],
                    self.licensing["identity_token"],
                    self.licensing["source_id"],
                )
                matlab_env["MLM_WEB_LICENSE"] = "true"
                matlab_env["MLM_WEB_USER_CRED"] = access_token_data["token"]
                matlab_env["MLM_WEB_ID"] = self.licensing["entitlement_id"]
                matlab_env["MW_LOGIN_EMAIL_ADDRESS"] = self.licensing["email_addr"]
                matlab_env["MW_LOGIN_FIRST_NAME"] = self.licensing["first_name"]
                matlab_env["MW_LOGIN_LAST_NAME"] = self.licensing["last_name"]
                matlab_env["MW_LOGIN_DISPLAY_NAME"] = self.licensing["display_name"]
                matlab_env["MW_LOGIN_USER_ID"] = self.licensing["user_id"]
                matlab_env["MW_LOGIN_PROFILE_ID"] = self.licensing["profile_id"]

                matlab_env["MHLM_CONTEXT"] = (
                    "MATLAB_JAVASCRIPT_DESKTOP"
                    if os.getenv(mwi_env.get_env_name_mhlm_context()) is None
                    else os.getenv(mwi_env.get_env_name_mhlm_context())
                )
            except OnlineLicensingError as e:
                raise e

        elif self.licensing["type"] == "nlm":
            matlab_env["MLM_LICENSE_FILE"] = self.licensing["conn_str"]

        # Env setup related to MATLAB
        matlab_env["MW_CRASH_MODE"] = "native"
        matlab_env["MATLAB_WORKER_CONFIG_ENABLE_LOCAL_PARCLUSTER"] = "true"
        matlab_env["PCT_ENABLED"] = "true"
        matlab_env["HTTP_MATLAB_CLIENT_GATEWAY_PUBLIC_PORT"] = "1"
        matlab_env["MW_DOCROOT"] = str(
            self.settings["matlab_path"] / "ui" / "webgui" / "src"
        )
        matlab_env["MWAPIKEY"] = self.settings["mwapikey"]

        # For r2020b, r2021a
        matlab_env["MW_CD_ANYWHERE_ENABLED"] = "true"
        # For >= r2021b
        matlab_env["MW_CD_ANYWHERE_DISABLED"] = "false"

        # DDUX info for MATLAB
        matlab_env["MW_CONTEXT_TAGS"] = self.settings.get("mw_context_tags")

        # Adding DISPLAY key which is only available after starting Xvfb successfully.
        matlab_env["DISPLAY"] = self.settings["matlab_display"]

        # MW_CONNECTOR_SECURE_PORT and MATLAB_LOG_DIR keys to matlab_env as they are available after
        # reserving port and preparing lockfiles for MATLAB
        matlab_env["MW_CONNECTOR_SECURE_PORT"] = str(self.matlab_port)

        # The matlab ready file is written into this location(self.matlab_ready_file_dir) by MATLAB
        # The matlab_ready_file_dir is where MATLAB will write any subsequent logs
        matlab_env["MATLAB_LOG_DIR"] = str(self.matlab_ready_file_dir)

        # Env setup related to logging
        # Very verbose logging in debug mode
        if logger.isEnabledFor(logging.getLevelName("DEBUG")):
            matlab_env["MW_DIAGNOSTIC_DEST"] = "stdout"
            matlab_env[
                "MW_DIAGNOSTIC_SPEC"
            ] = "connector::http::server=all;connector::lifecycle=all"

        # TODO Introduce a warmup flag to enable this?
        # matlab_env["CONNECTOR_CONFIGURABLE_WARMUP_TASKS"] = "warmup_hgweb"
        # matlab_env["CONNECTOR_WARMUP"] = "true"

        return matlab_env

    async def start_matlab(self, restart_matlab=False):
        """Start MATLAB.

        Args:
            restart_matlab (bool, optional): Whether to restart MATLAB. Defaults to False.

        Raises:
            Exception: When MATLAB is already running and restart is False.
            Exception: When MATLAB is not licensed.
        """

        # FIXME
        if await self.get_matlab_state() != "down" and restart_matlab is False:
            raise Exception("MATLAB already running/starting!")

        # FIXME
        if not self.is_licensed():
            raise Exception("MATLAB is not licensed!")

        if not self.is_matlab_present():
            self.error = MatlabInstallError("'matlab' executable not found in PATH")
            logger.error("'matlab' executable not found in PATH")
            self.logs["matlab"].clear()
            return

        # Ensure that previous processes are stopped
        await self.stop_matlab()

        # Clear MATLAB errors and logging
        self.error = None
        self.logs["matlab"].clear()

        try:
            # Start Xvfb process and update display number in settings
            create_xvfb_cmd = self.settings["create_xvfb_cmd"]
            xvfb_cmd, dpipe = create_xvfb_cmd()

            xvfb, display_port = await mw.create_xvfb_process(xvfb_cmd, dpipe)

            self.settings["matlab_display"] = ":" + str(display_port)
            self.processes["xvfb"] = xvfb
            logger.debug(f"Started Xvfb with PID={xvfb.pid} on DISPLAY={display_port}")

            # Finds and reserves a free port, then prepare lock files for the MATLAB process.
            self.prepare_lock_files_for_MATLAB_launch()

            # Configure the environment MATLAB needs to start
            matlab_env = await self.__setup_env_for_matlab()

        # If there's something wrong with setting up, capture the error for logging
        # and to pass to the front-end. Don't start the MATLAB process by returning early.
        except Exception as err:
            self.error = err
            log_error(logger, err)
            # stop_matlab() does the teardown work by removing any residual files and processes created till now.
            await self.stop_matlab()
            return

        # Start MATLAB Process
        logger.debug(f"Starting MATLAB on port {self.matlab_port}")
        master, slave = pty.openpty()
        matlab = await asyncio.create_subprocess_exec(
            *self.settings["matlab_cmd"],
            env=matlab_env,
            stdin=slave,
            stderr=asyncio.subprocess.PIPE,
        )
        self.processes["matlab"] = matlab
        logger.debug(f"Started MATLAB (PID={matlab.pid})")

        async def matlab_stderr_reader():
            while not self.processes["matlab"].stderr.at_eof():
                line = await self.processes["matlab"].stderr.readline()
                if line is None:
                    break
                self.logs["matlab"].append(line)
            await self.handle_matlab_output()

        loop = util.get_event_loop()

        self.tasks["matlab_stderr_reader"] = loop.create_task(matlab_stderr_reader())

    async def stop_matlab(self):
        """Terminate MATLAB."""

        # Cancel the asyncio task which reads MATLAB process' stderr
        if "matlab_stderr_reader" in self.tasks:
            try:
                self.tasks["matlab_stderr_reader"].cancel()
            except asyncio.CancelledError:
                pass

            del self.tasks["matlab_stderr_reader"]

        matlab = self.processes["matlab"]

        # Terminate
        if matlab is not None and matlab.returncode is None:
            # Instead of calling the .terminate() method first, which leads to MATLAB exiting with error code: 15,
            # try to shutdown MATLAB gracefully by sending the 'exit' command to the Embedded connector.
            # This will let MATLAB to safely checkin licenses and shutdown gracefully.
            # If the request fails, call the .terminate() method on the process.

            logger.info(f"Terminating MATLAB (PID={matlab.pid})")
            data = mwi.embedded_connector.helpers.get_data_to_eval_mcode("exit")
            url = mwi.embedded_connector.helpers.get_mvm_endpoint(
                self.settings["mwi_server_url"]
            )
            try:
                resp_json = await mwi.embedded_connector.send_request(
                    url=url, method="POST", data=data, headers=None
                )
                if resp_json["messages"]["EvalResponse"][0]["isError"]:
                    raise EmbeddedConnectorError(
                        "Failed to send HTTP request to Embedded connector"
                    )

                # Wait for matlab to shutdown gracefully
                await matlab.wait()
                assert (
                    matlab.returncode == 0
                ), "Failed to gracefully shutdown MATLAB via the embedded connector"

            except Exception as err:
                log_error(logger, err)
                # Additional try-catch as error could be thrown in Windows.
                try:
                    matlab.terminate()
                    await matlab.wait()
                except:
                    pass

        xvfb = self.processes["xvfb"]
        if xvfb is not None and xvfb.returncode is None:
            logger.info(f"Terminating Xvfb (PID={xvfb.pid})")
            xvfb.terminate()
            await xvfb.wait()

        try:
            # Clean up both connector.securePort file and mwi_proxy_lock_file
            if self.matlab_ready_file is not None:
                self.matlab_ready_file.unlink()

            if self.mwi_proxy_lock_file is not None:
                self.mwi_proxy_lock_file.unlink()

        except FileNotFoundError:
            # Files won't exist when stop_matlab is called for the first time.
            pass

        # Clear logs if MATLAB stopped intentionally
        self.logs["matlab"].clear()

    async def handle_matlab_output(self):
        """Parse MATLAB output from stdout and raise errors if any."""
        matlab = self.processes["matlab"]

        # Wait for MATLAB process to exit
        logger.info("Waiting for MATLAB to exit...")
        await matlab.wait()

        rc = self.processes["matlab"].returncode
        logger.info(
            f"MATLAB has shutdown with {'exit' if rc == 0 else 'error'} code: {rc}"
        )

        # Look for errors if MATLAB was not intentionally stopped and had an error code
        if len(self.logs["matlab"]) > 0 and rc != 0:
            err = None
            logs = [log.decode().rstrip() for log in self.logs["matlab"]]

            def parsed_errs():
                if self.licensing["type"] == "nlm":
                    yield mw.parse_nlm_error(logs, self.licensing["conn_str"])
                if self.licensing["type"] == "mhlm":
                    yield mw.parse_mhlm_error(logs)
                yield mw.parse_other_error(logs)

            for err in parsed_errs():
                if err is not None:
                    break

            if err is not None:
                self.error = err
                log_error(logger, err)



# Copyright 2020-2021 The MathWorks, Inc.

import asyncio
import os
import select
import xml.etree.ElementTree as ET

import aiohttp
from matlab_proxy.default_configuration import config
from matlab_proxy.util import mwi
from matlab_proxy.util.mwi.exceptions import (
    EntitlementError,
    MatlabError,
    NetworkLicensingError,
    OnlineLicensingError,
    XvfbError,
)

logger = mwi.logger.get()


def __get_licensing_url():
    """Private function to query for the licensing URL

    Returns:
        String: Licensing URL
    """
    return f"{config['doc_url']}blob/main/MATLAB-Licensing-Info.md"


async def fetch_entitlements(mhlm_api_endpoint, access_token, matlab_release):
    """Asynchronously fetch entitlements from MHLM endpoint. Used when licensing using MHLM.

    Args:
        mhlm_api_endpoint (String): URL of the API endpoint for fetching entitlements from MHLM.
        access_token (String): An access token which was requested by the fetch_acces_token() method
        matlab_release (String): MATLAB Release version installed in the system.
        mhlm_context: Environment specific info sent to the mhlm_api_endpoint

    Raises:
        OnlineLicensingError: Raised when unable to receive proper response from MHLM servers.
        EntitlementError: Raised when no entitlements are present for a license.

    Returns:
        list: Representing a list of Dicts containing the id, label and license_number.
    """
    # Get entitlements for token
    async with aiohttp.ClientSession() as client_session:
        async with client_session.post(
            mhlm_api_endpoint,
            headers={"content-type": "application/x-www-form-urlencoded"},
            data=aiohttp.FormData(
                {
                    "token": access_token,
                    "release": matlab_release,
                    "coreProduct": "ML",
                    "context": "jupyter",
                    "excludeExpired": "true",
                }
            ),
        ) as res:

            if res.reason != "OK":
                raise OnlineLicensingError(
                    f"Communication with {mhlm_api_endpoint} failed ({res.status}). For more details, see {__get_licensing_url()}."
                )

            root = ET.fromstring(await res.text())
            entitlement_el = root.find("entitlements")

            if entitlement_el is None or len(entitlement_el) == 0:
                raise EntitlementError(
                    f"Your MathWorks account is not linked to a valid license for MATLAB {matlab_release}."
                )

            entitlements = entitlement_el.findall("entitlement")

            return [
                {
                    "id": entitlement.find("id").text,
                    "label": entitlement.find("label").text,
                    "license_number": entitlement.find("license_number").text,
                }
                for entitlement in entitlements
            ]


async def fetch_expand_token(mwa_api_endpoint, identity_token, source_id):
    """Asynchronously fetch tokens from MWA API after MHLM licensing is successful.

    Args:
        mwa_api_endpoint (String): URL of the MWA API endpoint.
        identity_token (String): Identity token received from MHLM servers by the front end.
        source_id (String): Source ID received from MHLM servers by the front end.

    Raises:
        OnlineLicensingError: When unable to contact MWA API endpoint.

    Returns:
        Dict: Containing User and License expiration details.
    """
    async with aiohttp.ClientSession() as client_session:
        async with client_session.post(
            f"{mwa_api_endpoint}/tokens",
            headers={
                "content-type": "application/x-www-form-urlencoded",
                "accept": "application/json",
                "X_MW_WS_callerId": "desktop-jupyter",
            },
            data=aiohttp.FormData(
                {
                    "tokenString": identity_token,
                    "tokenPolicyName": "R2",
                    "sourceId": source_id,
                }
            ),
        ) as res:

            if res.reason != "OK":
                raise OnlineLicensingError(
                    f"Communication with {mwa_api_endpoint} failed ({res.status}). For more details, see {__get_licensing_url()}."
                )

            data = await res.json()

            return {
                "expiry": data["expirationDate"],
                "first_name": data["referenceDetail"]["firstName"],
                "last_name": data["referenceDetail"]["lastName"],
                "display_name": data["referenceDetail"]["displayName"],
                "user_id": data["referenceDetail"]["userId"],
                "profile_id": data["referenceDetail"]["referenceId"],
            }


async def fetch_access_token(mwa_api_endpoint, identity_token, source_id):
    """Asynchronously fetch access token from MWA API endpoint.

    Args:
        mwa_api_endpoint (String): URL of the MWA API endpoint.
        identity_token (String): String representing unique identity
        source_id (String): []

    Raises:
        OnlineLicensingError: When unable to contact MWA API endpoint.

    Returns:
        Dict : Containing the Access token.
    """
    async with aiohttp.ClientSession() as client_session:
        async with client_session.post(
            f"{mwa_api_endpoint}/tokens/access",
            headers={
                "content-type": "application/x-www-form-urlencoded",
                "accept": "application/json",
                "X_MW_WS_callerId": "desktop-jupyter",
            },
            data=aiohttp.FormData(
                {
                    "tokenString": identity_token,
                    "type": "MWAS",
                    "sourceId": source_id,
                }
            ),
        ) as res:

            if res.reason != "OK":
                raise OnlineLicensingError(
                    f"Communication with {mwa_api_endpoint} failed ({res.status}). For more details, see {__get_licensing_url()}."
                )

            data = await res.json()

            return {
                "token": data["accessTokenString"],
            }


def range_matlab_connector_ports():
    """Generator of acceptable ports for MATLAB Connector.
        Allowed ports conform to the regex: [3,6]1[5-9][1-9][1-9]

    Yields:
        Int : Representing a Port number.
    """
    for p1 in (3, 6):
        p2 = 1
        for p3 in range(5, 10):
            for p4 in range(1, 10):
                for p5 in range(1, 10):
                    yield int(f"{p1}{p2}{p3}{p4}{p5}")


def parse_nlm_error(logs, conn_str):
    """Parses error logs and returns NLM specific errors

    Args:
        logs (List): An array containing error logs.
        conn_str (String): The connection string of NLM.

    Returns:
        List: Containing NLM specific errors.
    """
    nlm_logs = []
    start = False
    for log in logs:
        if start is False:
            if "License checkout failed" in log:
                start = True
                nlm_logs.append(log)
        else:
            if "Diagnostic Information" in log:
                return NetworkLicensingError(
                    f"License checkout from {conn_str} failed. For more details, see {__get_licensing_url()}.",
                    logs=nlm_logs,
                )
            nlm_logs.append(log)
    return None


def parse_mhlm_error(logs):
    """Parses error logs and returns MHLM specific errors

    Args:
        logs (List): An array containing error logs.

    Returns:
        List: Containing MHLM specific errors.
    """
    mhlm_logs = None
    start = False
    for log in logs:
        if "License Manager Error" in log:
            start = True
            mhlm_logs = [log]

        if start is True:
            mhlm_logs.append(log)

    if mhlm_logs is not None:
        return OnlineLicensingError(
            f"Usage of MathWorks Online Licensing failed. For more details, see {__get_licensing_url()}.",
            logs=mhlm_logs,
        )
    return None


def parse_other_error(logs):
    """Returns a MATLAB specific error.

    Args:
        logs (List): An array containing error logs.

    Returns:
        MatlabError: An instance of MatlabError class with logs.
    """
    return MatlabError(
        "MATLAB returned an unexpected error. For more details, see the log below.",
        logs=logs,
    )


async def create_xvfb_process(xvfb_cmd, pipe, env={}):
    """Creates the Xvfb process.

    The Xvfb process is run with '-displayfd' flag set. This makes Xvfb choose an available
    display number and write it into the provided write descriptor. ie: pipe[1]

    We read this display number from the read descriptor.
    For this, we have 2 things to consider:
        1. How many bytes to read from the read descriptor. This is handled by the variable number_of_bytes.

        2. For how long to read from the read descriptor. We wait for atmost 10 seconds for Xvfb to write
           the display number using the 'select' package.


    Args:
        xvfb_cmd (List): A list containing the command to run the Xvfb process
        pipe (List): A list containing a pair of file descriptor.
        env (Dict): A Dict containing environment variables for the Xvfb process.

    Returns:
        List: Containing the Xvfb process object, and display number on which Xvfb process has started.
    """

    # Creates subprocess asynchronously with environment variables defined in matlab_env
    # Pipe errors, if any, to the process object instead of stdout.
    xvfb = await asyncio.create_subprocess_exec(
        *xvfb_cmd, close_fds=False, env=env, stderr=asyncio.subprocess.PIPE
    )

    read_descriptor, write_descriptor = pipe
    number_of_bytes = 200

    logger.debug("Waiting for XVFB process to initialize and provide Display Number")
    # Waits upto 10 seconds for the read_descriptor to be ready.
    ready_descriptors, _, _ = select.select([read_descriptor], [], [], 10)

    # If read_descriptor is in ready_descriptors, read from it.
    if read_descriptor in ready_descriptors:
        logger.debug("Reading display number from the read descriptor.")
        line = os.read(read_descriptor, number_of_bytes).decode("utf-8")
        # Xvfb process writes the display number and adds a new line character ('\n') at the end. Removing it with .strip()
        display_port = line.strip()

    else:
        # Check for errors and raise exception.
        error = ""
        while not xvfb.stderr.at_eof():
            line = await xvfb.stderr.readline()
            error += line.decode("utf-8")

        await xvfb.wait()
        raise XvfbError(f"Unable to start the Xvfb process: \n {error}")

    # Close the read and write descriptors.
    os.close(read_descriptor)
    os.close(write_descriptor)

    return xvfb, display_port



# Copyright 2021-2022 The MathWorks, Inc.

""" This file contains helper methods which return the details required for sending
a HTTP request to the Embedded Connector. """


from matlab_proxy.util import mwi
from matlab_proxy.util.mwi import environment_variables as mwi_env


def get_mvm_endpoint(mwi_server_url):
    """Returns the endpoint at which mvm can be communicated with

    Args:
        mwi_server_url (str): The complete origin where the matlab-proxy server
        is running at.
        Example: https://www.matlab-proxy.com:8056

    Returns:
        str: Complete origin URL to communicate with mvm
    """
    return f"{mwi_server_url}/messageservice/json/secure"


def get_ping_endpoint(mwi_server_url):
    """Returns the endpoint to communicate with MATLAB's Embedded Connector to
    get its latest status.

    Args:
        mwi_server_url (str): The complete origin where the matlab-proxy server
        is running at.
        Example: https://www.matlab-proxy.com:8056

    Returns:
        str: Complete origin URL to communicate with MATLAB's Embedded Connector.
    """
    return f"{mwi_server_url}/messageservice/json/state"


def get_data_for_ping_request():
    """Returns data required to send in the payload for a ping request to the embedded connector

    Returns:
        dict: Payload data
    """
    return {"messages": {"Ping": [{}]}}


def get_data_to_eval_mcode(m_code):
    """Returns the data required to send in the payload for evaluating mcode using eval function to the embedded connector.

    Args:
        m_code (str): MATLAB code to be evaluated

    Returns:
        dict: Payload data.
    """
    data = {
        "uuid": __generate_uuid(),
        "messages": {"Eval": [{"mcode": m_code, "uuid": __generate_uuid()}]},
        "computeToken": {"computeSessionId": "unused"},
    }

    return data


def get_data_to_feval_mcode(m_function, *args, nargout):
    """Returns the data required to send in the payload for evaluating mcode using Feval function to the embedded connector.

    Args:
        m_function (str): MATLAB function to be evaluated
        nargout (int): Number of return values the MATLAB function will return

    Returns:
        dict: Payload data.
    """
    data = {
        "uuid": __generate_uuid(),
        "messages": {
            "FEval": [
                {
                    "function": m_function,
                    "arguments": [*args],
                    "nargout": nargout,
                    "priority": 1,
                    "dequeMode": "ppe",
                    "uuid": __generate_uuid(),
                }
            ]
        },
        "computeToken": {"computeSessionId": "unused"},
    }

    return data


def __generate_uuid():
    """Generates random strings of length 8 containing mix of numbers and capital letters.
        Note: This function mimics the UUID generation in MOS.
    Returns:
        str: Generates a 8 character long string used to represent an UUID for use with JS API
    """
    import random

    uuid = ""
    numbers = "0123456789"
    letters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    for i in range(0, 8):
        if random.random() <= 0.5:
            index = random.randint(0, len(numbers) - 1)
            uuid += numbers[index]
        else:
            index = random.randint(0, len(letters) - 1)
            uuid += letters[index]

    return uuid



import pytest
from matlab_proxy.util import mwi
from matlab_proxy.util.mwi.exceptions import EmbeddedConnectorError
from tests.util import MockResponse


async def test_send_request_success(mocker):
    json_data = {"hello": "world"}

    payload = json_data
    mock_resp = MockResponse(payload=payload, reason="OK")

    mocked = mocker.patch("aiohttp.ClientSession.request", return_value=mock_resp)
    res = await mwi.embedded_connector.send_request(
        url="https://localhost:3000", data=json_data, method="GET"
    )

    assert json_data["hello"] == res["hello"]


async def test_send_request_failure(mocker):
    json_data = {"hello": "world"}

    payload = json_data
    mock_resp = MockResponse(payload=payload, reason="NOT-OK")

    mocked = mocker.patch("aiohttp.ClientSession.request", return_value=mock_resp)

    # Failed to communicate with EmbeddedConnector
    with pytest.raises(EmbeddedConnectorError):
        res = await mwi.embedded_connector.send_request(
            url="https://localhost:3000", method="GET", data=json_data
        )

    # Doesnt have url or data or method
    mock_resp = MockResponse(payload=payload, reason="NOT-OK")

    for key in ["url", "method"]:
        options = {
            "url": "https://localhost:3000",
            "data": json_data,
            "method": "GET",
        }
        options[key] = ""
        with pytest.raises(EmbeddedConnectorError):
            res = await mwi.embedded_connector.send_request(**options)



# Copyright 2021 The MathWorks, Inc.

import datetime
import random
import re
import secrets
from collections import namedtuple
from datetime import timedelta, timezone

import pytest
from matlab_proxy import settings
from matlab_proxy.util import mw
from matlab_proxy.util.mwi import exceptions

"""This file tests methods present in matlab_proxy/util/mw.py
"""


@pytest.fixture(name="mwa_api_data")
def mwa_api_data_fixture():
    """Pytest fixture which returns a namedtuple.

    The namedtuple contains values required for MW authentication

    Returns:
        namedtuple: A named tuple containing mwa, mhlm end-point URLs (with regex patterns), source_id, identity_token, access_token and matlab_release.
    """

    mwa_api_endpoint = "https://login.mathworks.com/authenticationws/service/v4"
    mwa_api_endpoint_pattern = re.compile("^" + mwa_api_endpoint)
    mhlm_api_endpoint = (
        "https://licensing.mathworks.com/mls/service/v1/entitlement/list"
    )
    mhlm_api_endpoint_pattern = re.compile("^" + mhlm_api_endpoint)
    mhlm_context = "jupyter"

    identity_token = secrets.token_urlsafe(324)
    source_id = secrets.token_urlsafe(21)
    access_token = secrets.token_urlsafe(22)
    matlab_release = "R2020b"

    mwa_api_variables = namedtuple(
        "mwa_api_variables",
        [
            "mwa_api_endpoint",
            "mwa_api_endpoint_pattern",
            "mhlm_api_endpoint",
            "mhlm_api_endpoint_pattern",
            "identity_token",
            "source_id",
            "access_token",
            "matlab_release",
            "mhlm_context",
        ],
    )

    variables = mwa_api_variables(
        mwa_api_endpoint,
        mwa_api_endpoint_pattern,
        mhlm_api_endpoint,
        mhlm_api_endpoint_pattern,
        identity_token,
        source_id,
        access_token,
        matlab_release,
        mhlm_context,
    )

    return variables


@pytest.fixture(name="fetch_access_token_valid_json")
def fetch_access_token_valid_json_fixture():
    """Pytest fixture which returns a dict.

    This fixture returns a dict representing a valid json response from mhlm servers.

    Returns:
        dict : A dictionary containing Key-value pairs present in a valid json response from mhlm servers.
    """

    now = datetime.datetime.now(timezone.utc)
    authentication_date = str(now.strftime("%Y-%m-%dT%H:%M:%S.%f%z"))
    expiration_date = str((now + timedelta(days=1)).strftime("%Y-%m-%dT%H:%M:%S.%f%z"))

    login_identifier = "abc@mathworks.com"
    reference_ID = int("".join([str(random.randint(0, 10)) for _ in range(8)]))

    first_name = "abc"
    last_name = "def"
    user_id = int("".join([str(random.randint(0, 10)) for _ in range(13)]))

    access_token_string = int("".join([str(random.randint(0, 10)) for _ in range(272)]))

    json_data = {
        "authenticationDate": authentication_date,
        "expirationDate": expiration_date,
        "id": 0,
        "loginIdentifier": login_identifier,
        "loginIdentifierType": "MWAS",
        "referenceDetail": {
            "referenceId": str(reference_ID),
            "country": "IN",
            "email": login_identifier,
            "firstName": first_name,
            "lastName": last_name,
            "displayName": first_name,
            "sector": "None",
            "userId": "mwa" + str(user_id),
            "profilePicture": "https://www.mathworks.com/",
        },
        "referenceId": str(reference_ID),
        "referenceType": "WEBPROFILEID",
        "source": "desktop-jupyter",
        "accessTokenString": str(access_token_string),
    }

    return json_data


class MockResponse:
    def __init__(self, reason, payload={}, status=200, text=""):
        self._payload = payload
        self._text = text
        self.reason = reason
        self.status = status

    async def json(self):
        return self._payload

    async def text(self):
        return self._text

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        pass


async def test_fetch_access_token(mwa_api_data, fetch_access_token_valid_json, mocker):
    """Test to check mw.fetch_access_token method returns valid json response.

    The mock_response fixture mocks the aiohttp.ClientSession().post() method to return a custom HTTP response.

    Args:
        mwa_api_data (namedtuple): A pytest fixture which returns a namedtuple containing values for MW servers.
        fetch_access_token_valid_json (Dict): A pytest fixture which returns a dict representing a valid json response.
        mocker: Built in pytest fixture which can be used to mock functions.
    """
    json_data = fetch_access_token_valid_json

    payload = dict(accessTokenString=json_data["accessTokenString"])

    mock_resp = MockResponse(payload=payload, reason="OK")

    url_pattern = mwa_api_data.mwa_api_endpoint_pattern

    mocked = mocker.patch("aiohttp.ClientSession.post", return_value=mock_resp)
    res = await mw.fetch_access_token(
        mwa_api_data.mwa_api_endpoint,
        mwa_api_data.identity_token,
        mwa_api_data.source_id,
    )

    _, args, _ = mocked.mock_calls[0]
    url = args[0]

    assert re.match(url_pattern, url)
    assert json_data["accessTokenString"] == res["token"]


async def test_fetch_access_token_licensing_error(mwa_api_data, mocker):
    """Test to check mw.fetch_access_token() method raises a mwi_exceptions.OnlineLicensingError.

    When an invalid response is received from the server, this test checks if mwi_exceptions.OnlineLicensingError is raised.
    Args:
        mocker: Built in pytest fixture which can be used to mock functions.
        mwa_api_data (namedtuple): A pytest fixture which returns a namedtuple containing values for MW authentication
    """

    url_pattern = mwa_api_data.mwa_api_endpoint_pattern

    mock_resp = MockResponse(payload={}, reason="NOT-OK", status=404)

    mocked = mocker.patch("aiohttp.ClientSession.post", return_value=mock_resp)

    with pytest.raises(exceptions.OnlineLicensingError):
        resp = await mw.fetch_access_token(
            mwa_api_data.mwa_api_endpoint,
            mwa_api_data.identity_token,
            mwa_api_data.source_id,
        )

    _, args, _ = mocked.mock_calls[0]
    url = args[0]
    assert re.match(url_pattern, url)


async def test_fetch_expand_token_licensing_error(mocker, mwa_api_data):
    """Test to check fetch_expand_token raises mwi_exceptions.OnlineLicensing error.

    Args:
        mocker: Built in pytest fixture which can be used to mock functions.
        mwa_api_data (namedtuple): A pytest fixture which returns a namedtuple containing values for MW authentication
    """
    url_pattern = mwa_api_data.mwa_api_endpoint_pattern
    mock_resp = MockResponse(payload={}, reason="NOT-OK", status=503)
    mocked = mocker.patch("aiohttp.ClientSession.post", return_value=mock_resp)

    with pytest.raises(exceptions.OnlineLicensingError):
        resp = await mw.fetch_expand_token(
            mwa_api_data.mwa_api_endpoint,
            mwa_api_data.identity_token,
            mwa_api_data.source_id,
        )

    _, args, _ = mocked.mock_calls[0]
    url = args[0]
    assert re.match(url_pattern, url)


@pytest.fixture(name="fetch_expand_token_valid_json")
def fetch_expand_token_valid_json_fixture():
    """Pytest fixture which returns a dict

    The return value represents a valid json response when mw.fetch_expand_token function is called.

    Returns:
        dict: A dict representing valid json response.
    """
    now = datetime.datetime.now(timezone.utc)
    expiration_date = str((now + timedelta(days=1)).strftime("%Y-%m-%dT%H:%M:%S.%f%z"))
    first_name = "abc"
    display_name = first_name
    last_name = "def"
    reference_ID = int("".join([str(random.randint(0, 10)) for _ in range(8)]))
    user_id = int("".join([str(random.randint(0, 10)) for _ in range(13)]))

    json_data = {
        "expirationDate": str(expiration_date),
        "referenceDetail": {
            "referenceId": str(reference_ID),
            "firstName": first_name,
            "lastName": last_name,
            "displayName": first_name,
            "userId": "mwa" + str(user_id),
        },
    }

    return json_data


async def test_fetch_expand_token(mocker, fetch_expand_token_valid_json, mwa_api_data):
    """Test to check if mw.fetch_expand_token returns a correct json response

    mock_response is used to mock ClientSession.post method to return a HTTP Response containing a valid json response.
    Args:
        mocker: Built in pytest fixture which can be used to mock functions.
        fetch_expand_token_valid_json (namedtuple): Pytest fixture which returns a dict which is returned by the server when no exception is raised.
        mwa_api_data (namedtuple): A namedtuple which contains info related to mwa.
    """
    json_data = fetch_expand_token_valid_json

    url_pattern = mwa_api_data.mwa_api_endpoint_pattern
    referenceDetail = dict(
        firstName=json_data["referenceDetail"]["firstName"],
        lastName=json_data["referenceDetail"]["lastName"],
        displayName=json_data["referenceDetail"]["displayName"],
        userId=json_data["referenceDetail"]["userId"],
        referenceId=json_data["referenceDetail"]["referenceId"],
    )

    payload = dict(
        expirationDate=json_data["expirationDate"], referenceDetail=referenceDetail
    )

    mock_resp = MockResponse(payload=payload, reason="OK", status=200)
    mocked = mocker.patch("aiohttp.ClientSession.post", return_value=mock_resp)

    resp = await mw.fetch_expand_token(
        mwa_api_data.mwa_api_endpoint,
        mwa_api_data.identity_token,
        mwa_api_data.source_id,
    )

    _, args, _ = mocked.mock_calls[0]
    url = args[0]

    assert resp is not None and len(resp.keys()) > 0
    assert re.match(url_pattern, url)


async def test_fetch_entitlements_licensing_error(mocker, mwa_api_data):
    """Test to check if fetch_entitlements raises mwi_exceptions.OnlineLicensingError.

    When an invalid response is received, this test checks if mwi_exceptions.OnlineLicenseError is raised.

    Args:
        mocker: Built in pytest fixture which can be used to mock functions.
        mwa_api_data (namedtuple): A namedtuple which contains info related to mwa.
    """
    url_pattern = mwa_api_data.mhlm_api_endpoint_pattern
    mock_resp = MockResponse(payload={}, reason="NOT-OK", status=503)
    mocked = mocker.patch("aiohttp.ClientSession.post", return_value=mock_resp)

    with pytest.raises(exceptions.OnlineLicensingError):
        resp = await mw.fetch_entitlements(
            mwa_api_data.mhlm_api_endpoint,
            mwa_api_data.access_token,
            mwa_api_data.matlab_release,
        )

    _, args, _ = mocked.mock_calls[0]
    url = args[0]
    assert re.match(url_pattern, url)


@pytest.fixture(
    name="invalid_entitlements",
    params=[
        """<?xml version="1.0" encoding="UTF-8"?>
                        <describe_entitlements_response>
                        </describe_entitlements_response>""",
        """<?xml version="1.0" encoding="UTF-8"?>
                        <describe_entitlements_response>
                            <entitlements>
                            </entitlements>
                        </describe_entitlements_response>""",
    ],
    ids=[
        "Invalid Entitlement : No entitlements tag",
        "Invalid Entitlement : Empty entitlements tag",
    ],
)
def invalid_entitlements_fixture(request):
    """A parameterized pytest fixture which returns invalid entitlements.


    Args:
        request : Built-in pytest fixture

    Returns:
        [String]: A string containing invalid Entitlements.
    """
    return request.param


async def test_fetch_entitlements_entitlement_error(
    mocker, mwa_api_data, invalid_entitlements
):
    """Test to check fetch_entitlements raises mwi_exceptions.EntitlementError.


    When invalid entitlements are received as a response, this test checks if mw.fetch_entitlements raises an
    mwi_exceptions.EntitlementError. mock_response mocks aiohttp.ClientSession.post method to send invalid entitlements as a HTTP response.

    Args:

        mocker: Built in pytest fixture which can be used to mock functions.
        mwa_api_data (namedtuple): A namedtuple which contains info related to mwa.
        invalid_entitlements (String): String containing invalid entitlements
    """
    url_pattern = mwa_api_data.mhlm_api_endpoint_pattern

    mock_resp = MockResponse(
        payload={}, reason="OK", text=invalid_entitlements, status=404
    )
    mocked = mocker.patch("aiohttp.ClientSession.post", return_value=mock_resp)

    with pytest.raises(exceptions.EntitlementError):
        resp = await mw.fetch_entitlements(
            mwa_api_data.mhlm_api_endpoint,
            mwa_api_data.access_token,
            mwa_api_data.matlab_release,
        )

    _, args, _ = mocked.mock_calls[0]
    url = args[0]
    assert re.match(url_pattern, url)


@pytest.fixture(name="valid_entitlements")
def valid_entitlements_fixture():
    """
    Pytest fixture which returns a string representing valid entitlements
    """
    id = int("".join([str(random.randint(0, 10)) for _ in range(7)]))
    license_number = int("".join([str(random.randint(0, 10)) for _ in range(8)]))

    return """<?xml version="1.0" encoding="UTF-8"?>
                <describe_entitlements_response>
                <entitlements>
                    <entitlement>
                        <id>%d</id>
                        <label>MATLAB</label>
                        <license_number>%d</license_number>
                        </entitlement>
                </entitlements>
                </describe_entitlements_response>""" % (
        id,
        license_number,
    )


async def test_fetch_entitlements(mocker, mwa_api_data, valid_entitlements):
    """Test to check test_fetch_entitlements returns valid response.


    mock_response mocks aiohttpClientSession.post() method  to return valid entitlements as a HTTP response
    Args:
        mocker: Built in pytest fixture which can be used to mock functions.
        mwa_api_data (namedtuple): A namedtuple which contains info related to mwa.
        valid_entitlements (String): String containing valid entitlements as a response.
    """

    url_pattern = mwa_api_data.mhlm_api_endpoint_pattern

    mock_resp = MockResponse(
        payload={}, reason="OK", text=valid_entitlements, status=404
    )
    mocked = mocker.patch("aiohttp.ClientSession.post", return_value=mock_resp)

    resp = await mw.fetch_entitlements(
        mwa_api_data.mhlm_api_endpoint,
        mwa_api_data.access_token,
        mwa_api_data.matlab_release,
    )

    _, args, _ = mocked.mock_calls[0]
    url = args[0]

    assert resp is not None and len(resp) > 0
    assert re.match(url_pattern, url)


def test_parse_mhlm_no_error():
    """Test to check mw.parse_mhlm_error() method returns none when no mhlm specific error
    is present in the logs.
    """
    logs = ["Starting MATLAB proxy-app", "Error parsing config, resetting."]

    actual_output = mw.parse_mhlm_error(logs)
    expected_output = None

    assert actual_output == expected_output


def test_parse_mhlm_error():
    """Test to check mw.parse_mhlm_error() returns an mwi_exceptions.OnlineLiceningError.

    When logs contain mhlm specific error information, this test checks if mwi_exceptions.OnlineLicensingError is raised.
    """
    logs = ["License Manager Error", "MHLM Licensing Failed"]
    actual_output = mw.parse_mhlm_error(logs)
    expected_output = exceptions.OnlineLicensingError

    assert isinstance(actual_output, expected_output)


def test_parse_nlm_no_error():
    """Test to check parse_nlm_error returns none when no nlm specific error information is present in logs."""
    logs = []
    conn_str = ""
    expected_output = None

    assert mw.parse_nlm_error(logs, conn_str) == expected_output


def test_parse_nlm_error():
    """Test to check parse_nlm_error() method returns an exception when logs contain an error.

    When logs contain nlm specific errors, this test checks if parse_nlm_error() raises mwi_exceptions.NetworkLicensingError.
    """
    logs = [
        "Starting MATLAB proxy-app",
        "License checkout failed",
        "Error parsing config, resetting.",
        "Diagnostic Information",
    ]

    conn_str = "abc@nlm"

    actual_output = mw.parse_nlm_error(logs, conn_str)
    expected_output = exceptions.NetworkLicensingError

    assert isinstance(actual_output, expected_output)


def test_parse_other_error():
    """This test checks if exception.MatlabError is raised when matlab processes returncode is not 0 and logs contain
    matlab specific information
    """
    logs = ["Starting MATLAB proxy-app", "Error parsing config, resetting."]

    expected_output = exceptions.MatlabError
    actual_output = mw.parse_other_error(logs)

    assert isinstance(actual_output, expected_output)


def test_range_matlab_connector_ports():
    """This test checks if the generator mw.range _matlab_connector_ports()
    yields consecutive port numbers.
    """
    port_range = mw.range_matlab_connector_ports()
    first_port = next(port_range)
    second_port = next(port_range)

    assert first_port + 1 == second_port


async def test_create_xvfb_process(loop):
    """Test to check if more than 1 xvfb process can be created with -displayfd flag

    Creates 2 xvfb processes with '-displayfd' flag and  checks if the processes are
    running on unique display ports
    """
    # Get command to launch xvfb with -displayfd flag.
    settings_1 = settings.get(dev=True)
    settings_2 = settings.get(dev=True)

    xvfb_cmd_1, pipe_1 = settings.create_xvfb_cmd()
    xvfb_cmd_2, pipe_2 = settings.create_xvfb_cmd()

    # Create Xvfb processes
    xvfb_1, display_port_1 = await mw.create_xvfb_process(xvfb_cmd_1, pipe_1, {})
    xvfb_2, display_port_2 = await mw.create_xvfb_process(xvfb_cmd_2, pipe_2, {})

    # Verify
    assert xvfb_1 is not None and xvfb_2 is not None
    assert display_port_1 != display_port_2

    # Clean up
    xvfb_1.terminate()
    await xvfb_1.wait()

    xvfb_2.terminate()
    await xvfb_2.wait()



# Copyright 2020-2021 The MathWorks, Inc.

import json
import os
import stat
import time
from json.decoder import JSONDecodeError

import pytest
from matlab_proxy.util.mwi import custom_http_headers
from matlab_proxy.util.mwi import environment_variables as mwi_env


def test_get_custom_header_env_var():
    """Test to check if the __get_custom_header_env_var() method returns the expected environment variable name"""
    assert (
        mwi_env.get_env_name_custom_http_headers()
        == custom_http_headers.__get_custom_header_env_var()
    )


def test_get_exception_statement():
    """Test to check if __get_exception_statement() contains 'JSON data' in the generic exception statement returned by __get_exception_statement()"""
    assert "JSON data" in custom_http_headers.__get_exception_statement()


@pytest.fixture(name="non_existent_temp_json_file")
def non_existent_random_json_file_fixture(tmp_path):
    """Pytest fixture which returns a non-existent random json file within a temporary directory

    Args:
        tmp_path : Built in Pytest fixture.

    Returns:
        PosixPath: of the non-existent random json file.
    """
    random_file = tmp_path / f'{str(time.time()).replace(".", "")}.json'
    return random_file


@pytest.fixture(name="monkeypatch_env_var_with_json_file")
def monkeypatch_env_var_with_json_file_fixture(
    monkeypatch, non_existent_temp_json_file
):
    """Pytest fixture which monkeypatches the env variable returned by the __get_custom_header_env_var()
    method with the value returned by the pytest fixture non_existent_tmp_json_file

    Args:
        monkeypatch : Built in pytest fixture
        non_existent_temp_json_file: Pytest fixture which returns a string containing path
                                        to non-existent file in a temporary directory.
    """
    monkeypatch.setenv(
        mwi_env.get_env_name_custom_http_headers(),
        str(non_existent_temp_json_file),
    )


@pytest.fixture(name="valid_json_content")
def valid_json_content_fixture():
    """Pytest fixture which returns valid JSON data as a string

    Returns:
        [String]: containing valid JSON data.
    """
    return '{"abcd": "hello"}'


@pytest.fixture(name="json_file_with_valid_json")
def json_file_with_valid_json_fixture(non_existent_temp_json_file, valid_json_content):
    """Pytest fixture which returns a random json file containing valid json data

    Args:
        non_existent_temp_json_file : Pytest fixture which returns a non-existent random json file.
        valid_json_content : Pytest fixture which returns valid JSON data as a string.

    Returns:
        [PosixPath]: Of a random json file containing valid json data.
    """
    json_file_with_valid_content = non_existent_temp_json_file
    with open(json_file_with_valid_content, "w") as f:
        f.write(valid_json_content)

    return json_file_with_valid_content


def test_get_file_contents_with_valid_json(json_file_with_valid_json):
    """Test to check if __get_file_contents() returns valid json data as a dict from a file

    Args:
        json_file_with_valid_json : Pytest fixture which returns a non-existent random json file.
    """
    file_content = None
    with open(json_file_with_valid_json, "r") as f:
        file_content = json.load(f)

    assert file_content == custom_http_headers.__get_file_contents(
        json_file_with_valid_json
    )


@pytest.fixture(name="invalid_json_content")
def invalid_json_content_fixture():
    """Pytest fixture which returns invalid json data as a string

    Returns:
        [String]: containing invalid json data.
    """
    return '{"abcd"= "hello"}'


@pytest.fixture(name="json_file_with_invalid_json")
def json_file_with_invalid_json_fixture(
    non_existent_temp_json_file, invalid_json_content
):
    """Pytest fixture which returns a path to a random json file containing invalid json data.

    Args:
        non_existent_temp_json_file : Pytest fixture which returns a non-existent random json file.
        invalid_json_content : Pytest fixture which returns invalid json data as a string.

    Returns:
        PosixPath: to a random file containing invalid json data.
    """
    json_file_with_invalid_content = non_existent_temp_json_file
    with open(json_file_with_invalid_content, "w") as f:
        f.write(invalid_json_content)

    return json_file_with_invalid_content


def test_get_file_contents_with_invalid_json(json_file_with_invalid_json, capsys):
    """Test to check __get_file_contents() raises JSONDecodeError when invalid json data is present in a file.

    Args:
        json_file_with_invalid_json : Pytest fixture which returns a non-existent random json file.
    """
    with pytest.raises(SystemExit):
        custom_http_headers.__get_file_contents(json_file_with_invalid_json)
        out, err = capsys.readouterr()
        assert JSONDecodeError.__name__ in out


def test_check_file_validity_no_read_access(non_existent_temp_json_file, capsys):
    """Test to check if OSError is raised when trying to read file with no read access.

    Args:
        non_existent_temp_json_file : Pytest fixture which returns a non-existent random json file.
    """
    temp_json_file_no_read_access = non_existent_temp_json_file
    temp_json_file_no_read_access.touch(mode=stat.S_IWUSR)

    with pytest.raises(SystemExit):
        custom_http_headers.__check_file_validity(temp_json_file_no_read_access)
        out, err = capsys.readouterr()
        assert OSError.__name__ in out


def test_check_file_validity_no_error(non_existent_temp_json_file, valid_json_content):
    """Test to check __check_file_validity() does not raise any exception when the custom headers file
    exists, has valid JSON data and the current python process has read access to the file.

    Args:
        non_existent_temp_json_file : Pytest fixture which returns a non-existent random json file.
    """

    random_json_file_with_valid_content = non_existent_temp_json_file
    with open(random_json_file_with_valid_content, "w") as f:
        f.write(valid_json_content)

    assert (
        custom_http_headers.__check_file_validity(random_json_file_with_valid_content)
        is True
    )


def test_get_no_env_var():
    """Test to check if get() returns an empty dict, when mwi_custom_http_headers env variable is not present."""
    assert custom_http_headers.get() == dict()


def test_get_with_json_file_no_error(
    monkeypatch_env_var_with_json_file, valid_json_content
):
    """Test to check if expected json content is returned by get() as a dict when the
    environment variable returned by __get_custom_header_env_var() contains a path to a JSON file

    Args:
        monkeypatch_env_var_with_json_file : Pytest fixture which monkeypatches the env variable returned by __get_custom_header_env_var() to
                              have a non-existent temporary json file.
        valid_json_content : Pytest fixture which returns valid json data as a string.
    """
    tmp_file_path = os.getenv(mwi_env.get_env_name_custom_http_headers())

    with open(tmp_file_path, "w") as f:
        f.write(valid_json_content)

    assert custom_http_headers.get() == json.loads(valid_json_content)


def test_get_with_json_file_raise_exception(
    monkeypatch_env_var_with_json_file, invalid_json_content, capsys
):
    """Test to check if the get() method raises SystemExit exception when the environment variable returned by __get_custom_env_var()
    contains path to a file with invalid JSON data. Also asserts for JSONDecodeError in stdout.

    Args:
        monkeypatch_env_var_with_json_file : Pytest fixture which monkeypatches the env variable returned by __get_custom_header_env_var() to
                              have a non-existent temporary json file.
        invalid_json_content : Pytest fixture which returns invalid json data as a string.
    """
    tmp_file_path = os.getenv(mwi_env.get_env_name_custom_http_headers())

    with open(tmp_file_path, "w") as f:
        f.write(invalid_json_content)

    with pytest.raises(SystemExit):
        custom_http_headers.get()
        out, err = capsys.readouterr()
        assert JSONDecodeError.__name__ in out


@pytest.fixture(name="monkeypatch_env_var_with_invalid_json_string")
def monkeypatch_env_var_with_invalid_json_string_fixture(
    monkeypatch, invalid_json_content
):
    """Pytest fixture which monkeypatches the env variable returned by the __get_custom_header_env_var()
    method with the value returned by the pytest fixture non_existent_temp_json_file

    Args:
        monkeypatch : Built in pytest fixture
        non_existent_temp_json_file: Pytest fixture which returns a string containing path
                                        to non-existent file in a temporary directory.
    """
    monkeypatch.setenv(mwi_env.get_env_name_custom_http_headers(), invalid_json_content)


def test_get_with_invalid_json_string(
    monkeypatch_env_var_with_invalid_json_string, capsys
):
    """Test to check if the get() method raises SystemExit exception when the environment variable returned by __get_custom_env_var()
    contains invalid JSON data as a string. Also asserts for JSONDecodeError in stdout.

    Args:
        monkeypatch_env_var_with_invalid_json_string : Pytest fixture which monkeypatches the env variable returned by __get_custom_header_env_var() to
                                                     contain invalid JSON data as a string.
    """
    with pytest.raises(SystemExit):
        headers = custom_http_headers.get()
        out, err = capsys.readouterr()
        assert JSONDecodeError.__name__ in out


@pytest.fixture(name="monkeypatch_env_var_with_valid_json_string")
def monkeypatch_env_var_with_valid_json_string_fixture(monkeypatch, valid_json_content):
    """Pytest fixture which monkeypatches the env variable returned by the __get_custom_header_env_var()
    method with the value returned by the pytest fixture non_existent_tmp_json_file

    Args:
        monkeypatch : Built in pytest fixture
        non_existent_temp_json_file: Pytest fixture which returns a string containing path
                                        to non-existent file in a temporary directory.
    """
    monkeypatch.setenv(mwi_env.get_env_name_custom_http_headers(), valid_json_content)


def test_get_with_valid_json_string(monkeypatch_env_var_with_valid_json_string):
    """Test to check if expected json content is returned by get() as a dict when the
    environment variable contains valid JSON data as a string

    Args:
        monkeypatch_env_var_with_valid_json_string : Pytest fixture which monkeypatches the env variable returned by __get_custom_header_env_var() to
                                                     contain valid JSON data as a string.
    """
    headers = json.loads(os.getenv(mwi_env.get_env_name_custom_http_headers()))
    assert headers == custom_http_headers.get()



# Copyright 2021 The MathWorks, Inc.
"""Tests for functions in matlab_proxy/util/mwi_validators.py
"""

import os
import random
import socket
import tempfile

import matlab_proxy
import pytest
from matlab_proxy.util.mwi import environment_variables as mwi_env
from matlab_proxy.util.mwi import validators
from matlab_proxy.util.mwi.exceptions import NetworkLicensingError


def test_validate_mlm_license_file_for_invalid_string(monkeypatch):
    """Check if validator raises expected exception"""
    # Delete the environment variables if they do exist
    env_name = mwi_env.get_env_name_network_license_manager()
    invalid_string = "/Invalid/String/"
    monkeypatch.setenv(env_name, invalid_string)
    nlm_conn_str = os.getenv(env_name)
    with pytest.raises(NetworkLicensingError) as e_info:
        conn_str = validators.validate_mlm_license_file(nlm_conn_str)
    assert invalid_string in str(e_info.value)


def test_validate_mlm_license_file_for_valid_server_syntax(monkeypatch):
    """Check if port@hostname passes validation"""
    env_name = mwi_env.get_env_name_network_license_manager()
    license_manager_address = "1234@1.2_any-alphanumeric"
    monkeypatch.setenv(env_name, license_manager_address)
    conn_str = validators.validate_mlm_license_file(os.getenv(env_name))
    assert conn_str == license_manager_address


def test_validate_mlm_license_file_for_valid_server_triad_syntax(monkeypatch):
    """Check if port@hostname passes validation"""
    env_name = mwi_env.get_env_name_network_license_manager()
    license_manager_address = (
        "1234@1.2_any-alphanumeric,1234@1.2_any-alphanumeric,1234@1.2_any-alphanumeric"
    )
    monkeypatch.setenv(env_name, license_manager_address)
    conn_str = validators.validate_mlm_license_file(os.getenv(env_name))
    assert conn_str == license_manager_address


def test_validate_mlm_license_file_None():
    """Test to check if validate_mlm_license_file() returns None when nlm_conn_str is None."""
    assert validators.validate_mlm_license_file(None) is None


def test_get_with_environment_variables(monkeypatch):
    """Check if path to license file passes validation"""
    env_name = mwi_env.get_env_name_network_license_manager()
    fd, path = tempfile.mkstemp()
    monkeypatch.setenv(env_name, path)
    try:
        conn_str = validators.validate_mlm_license_file(os.getenv(env_name))
        assert conn_str == str(path)
    finally:
        os.remove(path)


def test_validate_app_port_is_free_false():
    """Test to validate if supplied app port is free"""
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    port = s.getsockname()[1]
    with pytest.raises(SystemExit) as e:
        validators.validate_app_port_is_free(port)
    assert e.value.code == 1
    s.close()


def test_validate_app_port_is_free_true():
    """Test to validate if supplied app port is free"""
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    port = s.getsockname()[1]
    s.close()
    assert validators.validate_app_port_is_free(port) == port


def test_validate_app_port_None():
    """Tests if validated app port is None when MWI_APP_PORT env variable is not set.
    If validated app port is None implies a random free port will be used at launch.
    """
    assert validators.validate_app_port_is_free(None) is None


def test_validate_env_config_true():
    """Validate the default config which is used in this package."""
    config = validators.validate_env_config(matlab_proxy.get_default_config_name())
    assert isinstance(config, dict)


def test_validate_env_config_false():
    """Passing a non existent config should raise SystemExit exception"""

    with pytest.raises(SystemExit) as e:
        config = validators.validate_env_config(str(random.randint(10, 100)))

    assert e.value.code == 1


def test_get_configs():
    """Test to check if atleast 1 env config is discovered.
    When this package is installed, we will have a default config.
    """
    configs = validators.__get_configs()

    assert len(configs.keys()) >= 1


@pytest.mark.parametrize(
    "base_url, validated_base_url",
    [
        ("", ""),
        ("/bla", "/bla"),
        ("/bla/", "/bla"),
    ],
    ids=[
        "Launch integration at root",
        "Launch at custom path",
        "Launch at custom with suffix: /",
    ],
)
def test_validate_base_url(base_url, validated_base_url):
    """Tests multiple base_urls which will beparsed and validated successfully.

    Args:
        base_url (str): base_url
        validated_base_url (str): validated base_url
    """
    assert validators.validate_base_url(base_url) == validated_base_url


def test_validate_base_url_no_prefix_error():
    """Test to check base_url will throw error when a prefix / is not present in it.[summary]"""
    with pytest.raises(SystemExit) as e:
        validators.validate_base_url("matlab/")
    assert e.value.code == 1


def test_validate_mwi_ssl_key_and_cert_file(monkeypatch):
    """Check if port@hostname passes validation"""
    ssl_cert_file_env_name = mwi_env.get_env_name_ssl_cert_file()
    ssl_key_file_env_name = mwi_env.get_env_name_ssl_key_file()
    fd, path = tempfile.mkstemp()
    monkeypatch.setenv(ssl_cert_file_env_name, path)
    monkeypatch.setenv(ssl_key_file_env_name, path)
    try:
        # Verify that if KEY and CERT are provided
        key_file, cert_file = validators.validate_ssl_key_and_cert_file(
            os.getenv(ssl_key_file_env_name), os.getenv(ssl_cert_file_env_name)
        )
        assert key_file == str(path)
        assert cert_file == str(path)

        # Verify that KEY can be None
        key_file, cert_file = validators.validate_ssl_key_and_cert_file(
            None, os.getenv(ssl_cert_file_env_name)
        )
        assert key_file == None
        assert cert_file == str(path)

        # Verify that if KEY is provided, CERT must also be provided
        with pytest.raises(SystemExit) as e:
            validators.validate_ssl_key_and_cert_file(
                os.getenv(ssl_key_file_env_name), None
            )
        assert e.value.code == 1

        # Verify that KEY is valid file location
        with pytest.raises(SystemExit) as e:
            validators.validate_ssl_key_and_cert_file(
                "/file/does/not/exist", os.getenv(ssl_cert_file_env_name)
            )
        assert e.value.code == 1

        # Verify that KEY is valid file location
        with pytest.raises(SystemExit) as e:
            validators.validate_ssl_key_and_cert_file(
                os.getenv(ssl_key_file_env_name), "/file/does/not/exist"
            )
        assert e.value.code == 1
    finally:
        os.remove(path)



# Copyright 2020-2021 The MathWorks, Inc.


def get_entrypoint_name():
    """Returns the entry_point name which will be registered when installing the package.

    Returns:
        str: entry_point name which will be used to register when installing this package.
    """
    return "matlab_proxy_configs"


def get_default_config_name():
    """Returns the default config name which will be used to launch the integration

    Returns:
        Str: default config name
    """
    return "default_configuration_matlab_proxy"


def get_executable_name():
    """Returns the executable name used to launch the integration

    Returns:
        str: Name of the executable
    """
    return "matlab-proxy-app"


def __get_matlab_proxy_base_ddux_value():
    """Returns the DDUX value for MATLAB use when launched its by matlab-proxy

    Returns:
        str : DDUX value for MATLAB use.
    """

    return "MATLAB_PROXY:BASE:V1"


def get_mwi_ddux_value(extension_name):
    """Returns DDUX value for matlab-proxy based on the context from which
    it is being launched from.

    Args:
        extension_name (str): The name of the extension/environment

    Returns:
        str: DDUX value for matlab-proxy based on the environment.
    """
    matlab_proxy_ddux_value = __get_matlab_proxy_base_ddux_value()

    if extension_name == get_default_config_name():
        return matlab_proxy_ddux_value
    else:
        variant = extension_name.upper().strip()
        variant = variant.replace(" ", "_").replace("-", "_")
        mwi_ddux_value = matlab_proxy_ddux_value.replace("BASE", variant)
        return mwi_ddux_value



# Copyright 2021-2022 The MathWorks, Inc.

""" 
This file contains the methods to communicate with the embedded connector.
"""


import json

import aiohttp
from matlab_proxy.util.mwi.exceptions import EmbeddedConnectorError

from .helpers import get_data_for_ping_request, get_ping_endpoint


async def send_request(url: str, data: dict, method: str, headers: dict = None) -> dict:
    """A helper method to send various kinds of HTTP requests to the embedded connector.
    The url and method params are required.

    Args:
        url (str): URL to send HTTP request
        method (str): HTTP Request type.
        payload (dict): Payload for the HTTP request
        headers (dict): Headers for the HTTP request.

    Raises:
        EmbeddedConnectorError: When unable to get a response from the Embedded connector

    Returns:
        dict: The json response from Embedded connector
    """

    if not url or not method:
        raise EmbeddedConnectorError(
            f"url or method key is missing in the options parameter."
        )

    if data and isinstance(data, dict):
        data = json.dumps(data)

    try:
        async with aiohttp.ClientSession() as session:
            async with session.request(
                method=method, url=url, data=data, headers=None, ssl=False
            ) as resp:
                if resp.reason != "OK":

                    # Converting to dict and formatting for printing
                    data = json.loads(data)

                    raise EmbeddedConnectorError(
                        f"""Failed to communicate with Embedded Connector.\nHTTP Request details:\n{json.dumps(data, indent=2)}"""
                    )

                return await resp.json()
    except Exception as err:
        raise err


async def get_state(mwi_server_url):
    """Returns the state of MATLAB's Embedded Connector.

    Args:
        port (int): The port on which the embedded connector is running at

    Returns:
        str: Either "up" or "down"
    """
    data = get_data_for_ping_request()
    url = get_ping_endpoint(mwi_server_url)
    try:
        resp = await send_request(url=url, data=data, method="POST")
        if not resp["messages"]["PingResponse"][0]["messageFaults"]:
            return "up"
    except Exception:
        pass

    return "down"



# Copyright 2020-2021 The MathWorks, Inc.


class AppError(Exception):
    """A Generic Parent class which inherits the Exception class.
    This class will be inherited by other classes representing specific exceptions.

    The Parameterized constructor stores the message, logs and stacktrace.

    Args:
        Exception : Python's inbuilt Exception Class.
    """

    def __init__(self, message, logs=None, stacktrace=None):
        self.message = message
        self.logs = logs
        self.stacktrace = stacktrace


class InternalError(AppError):
    """A Class which inherits the AppError class.
    This class represents any Internal Error within the App

    Args:
        AppError (Class): Parent Class containing attributes to store
        messages, logs and stacktrace.
    """

    pass


class MatlabInstallError(AppError):
    """A Class which inherits the AppError class.

    This class represents errors with MATLAB Installation.

    Args:
        AppError (Class): Parent Class containing attributes to store
        messages, logs and stacktrace.
    """

    pass


class LicensingError(AppError):
    """A Class which inherits the AppError class.

    This class represents any Licensing Errors (MHLM and NLM Licensing)

    Args:
        AppError (Class): Parent Class containing attributes to store
        messages, logs and stacktrace.
    """

    pass


class OnlineLicensingError(LicensingError):
    """A Class which inherits the Licensing class.

    This class represents any errors specific to MHLM Licensing.

    Args:
        LicensingError (Class): Parent Class representing Licensing Errors.
    """

    pass


class EntitlementError(OnlineLicensingError):
    """A Class which inherits the OnlineLicensingError class.

    This class represents errors with Entitilments in MHLM Licensing.

    Args:
        OnlineLicensingError  (Class): Parent Class representing errors specific to MHLM Licensing.
    """

    pass


class NetworkLicensingError(LicensingError):
    """A Class which inherits the Licensing class.

    This class represents errors specific to Network License Manager.

    Args:
        LicensingError (Class): Parent Class representing Licensing Errors.
    """

    pass


class NoAvailableNetworkLicensingError(NetworkLicensingError):
    """A Class which inherits the NetworkLicensingError class.

    This class represents errors specific to non-availability of Network License Manager.

    Args:
        NetworkLicensingError (Class): Parent Class representing errors specific to Network Licensing.
    """

    pass


class MatlabError(AppError):
    """A Class which inherits the AppError class.

    This class represents errors raised by MATLAB.

    Args:
        AppError (Class): Parent Class containing attributes to store
        messages, logs and stacktrace.
    """

    pass


class XvfbError(AppError):
    """A Class which inherits the AppError class.

    This class represents any errors raised by Xvfb process.

    Args:
        AppError (Class): Parent Class containing attributes to store
        messages, logs and stacktrace.
    """

    pass


class EmbeddedConnectorError(MatlabError):
    """A Class which inherits the MatlabError class.

    This class represents errors raised when proxy fails to communicate with the Embedded Connector.

    Args:
        AppError (Class): Parent Class containing attributes to store
        messages, logs and stacktrace.
    """

    pass


def log_error(logger, err):
    """Logs any error to stdout.

    Args:
        logger (logging): A instance of the logging.getLogger()
        err (Class): An instance of one of the  Error classes as defined above.
        Example: OnlineLicensingError, EntitlementError
    """
    logs_str = ("\n" + "\n".join(err.logs)) if err.logs is not None else ""
    logger.error(err.message + logs_str)



# Copyright 2020-2022 The MathWorks, Inc.
import os
from setuptools.command.install import install
import setuptools
import matlab_proxy
from pathlib import Path
from shutil import which
from matlab_proxy.default_configuration import config

npm_install = ["npm", "--prefix", "gui", "install", "gui"]
npm_build = ["npm", "run", "--prefix", "gui", "build"]


class InstallNpm(install):
    def run(self):

        # Ensure npm is present
        if which("npm") is None:
            raise Exception(
                "npm must be installed and on the path during package install!"
            )

        self.spawn(npm_install)
        self.spawn(npm_build)
        target_dir = Path(self.build_lib) / matlab_proxy.__name__ / "gui"
        self.mkpath(str(target_dir))
        self.copy_tree("gui/build", str(target_dir))

        # In order to be accessible in the package, turn the built gui into modules
        (Path(target_dir) / "__init__.py").touch(exist_ok=True)
        for (path, directories, filenames) in os.walk(target_dir):
            for directory in directories:
                (Path(path) / directory / "__init__.py").touch(exist_ok=True)

        super().run()


tests_require = [
    "pytest",
    "pytest-env",
    "pytest-cov",
    "pytest-mock",
    "pytest-aiohttp",
    "requests",
    "psutil",
]

HERE = Path(__file__).parent.resolve()
long_description = (HERE / "README.md").read_text()


setuptools.setup(
    name="matlab-proxy",
    version="0.2.9",
    url=config["doc_url"],
    author="The MathWorks, Inc.",
    author_email="cloud@mathworks.com",
    license="MATHWORKS CLOUD REFERENCE ARCHITECTURE LICENSE",
    description="Python® package enables you to launch MATLAB® and access it from a web browser.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(exclude=["devel", "tests", "anaconda"]),
    keywords=[
        "Proxy",
        "MATLAB Proxy",
        "MATLAB",
        "MATLAB Javascript Desktop",
        "MATLAB Web Desktop",
        "Remote MATLAB Web Access",
    ],
    classifiers=[
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    python_requires="~=3.6",
    install_requires=["aiohttp>=3.7.4"],
    tests_require=tests_require,
    extras_require={
        "dev": ["aiohttp-devtools", "black", "ruamel.yaml"] + tests_require
    },
    # The entrypoint will be used by multiple packages that have this package as an installation
    # dependency. These packages can use the same API, get_entrypoint_name(), to make their configs discoverable
    entry_points={
        matlab_proxy.get_entrypoint_name(): [
            f"{matlab_proxy.get_default_config_name()} = matlab_proxy.default_configuration:config"
        ],
        "console_scripts": [
            f"{matlab_proxy.get_executable_name()} = matlab_proxy.app:main"
        ],
    },
    include_package_data=True,
    zip_safe=False,
    cmdclass={"install": InstallNpm},
)



# Copyright 2020-2021 The MathWorks, Inc.

import asyncio
import json
import mimetypes
import pkgutil
import sys

import aiohttp
from aiohttp import web

import matlab_proxy
from matlab_proxy import settings, util
from matlab_proxy.app_state import AppState
from matlab_proxy.default_configuration import config
from matlab_proxy.util import mwi
from matlab_proxy.util.mwi import environment_variables as mwi_env
from matlab_proxy.util.mwi.exceptions import LicensingError

mimetypes.add_type("font/woff", ".woff")
mimetypes.add_type("font/woff2", ".woff2")
mimetypes.add_type("font/eot", ".eot")
mimetypes.add_type("font/ttf", ".ttf")
mimetypes.add_type("application/json", ".map")
mimetypes.add_type("image/png", ".ico")

# TODO It is bad practice to have global state in aiohttp applications, instead this
# mount point should be read in the application start up function, then if it is not
# an empty string, registering a subapp with the given prefix. In addition, if it is
# possible and does not mess-up the MATLAB iframe base_url detection, register a
# nested subapp (with a name unlikely to cause collisions) containing this API and the
# static serving.

# Setup logger for the integration and a web logger. Override any default loggers.
logger = mwi.logger.get(init=True)


def marshal_licensing_info(licensing_info):
    """Gather/Marshal licensing information for MHLM and NLM Licensing types.

    Args:
        licensing_info (Dict): Contains licensing information

    Returns:
        Dict: Licensing information specific to the type of Licensing.
    """
    if licensing_info is None:
        return None

    if licensing_info["type"] == "mhlm":
        return {
            "type": "MHLM",
            "emailAddress": licensing_info["email_addr"],
            "entitlements": licensing_info.get("entitlements", []),
            "entitlementId": licensing_info.get("entitlement_id", None),
        }
    elif licensing_info["type"] == "nlm":
        return {
            "type": "NLM",
            "connectionString": licensing_info["conn_str"],
        }


def marshal_error(error):
    """Gather/Marshal error details.

    Args:
        error (Object): Instance of NetworkLicensingError or OnlineLicensingError or MatlabError

    Returns:
        Dict: Containing information about the error.
    """
    if error is None:
        return None
    return {
        "message": error.message,
        "logs": error.logs,
        "type": error.__class__.__name__,
    }


async def create_status_response(app, loadUrl=None):
    """Send a generic status response about the state of server,MATLAB and MATLAB Licensing

    Args:
        app (aiohttp.web.Application): Web Server
        loadUrl (String, optional): Represents the root URL. Defaults to None.

    Returns:
        JSONResponse: A JSONResponse object containing the generic state of the server, MATLAB and MATLAB Licensing.
    """
    state = app["state"]
    return web.json_response(
        {
            "matlab": {
                "status": await state.get_matlab_state(),
                "version": state.settings["matlab_version"],
            },
            "licensing": marshal_licensing_info(state.licensing),
            "loadUrl": loadUrl,
            "error": marshal_error(state.error),
            "wsEnv": state.settings["ws_env"],
        }
    )


async def get_env_config(req):
    """API Endpoint to get Matlab Web Desktop environment specific configuration.

    Args:
        req (HTTPRequest): HTTPRequest Object.

    Returns:
        JSONResponse: contains a Dict representing environment specific configuration serialized to JSON
    """
    config = req.app["state"].settings["env_config"]
    return web.json_response(config)


async def get_status(req):
    """API Endpoint to get the generic status of the server, MATLAB and MATLAB Licensing.

    Args:
        req (HTTPRequest): HTTPRequest Object.

    Returns:
        JSONResponse: JSONResponse object containing information about the server, MATLAB and MATLAB Licensing.
    """
    return await create_status_response(req.app)


async def start_matlab(req):
    """API Endpoint to start MATLAB

    Args:
        req (HTTPRequest): HTTPRequest Object

    Returns:
        JSONResponse: JSONResponse object containing updated information on the state of MATLAB among other information.
    """
    state = req.app["state"]

    # Start MATLAB
    await state.start_matlab(restart_matlab=True)

    return await create_status_response(req.app)


async def stop_matlab(req):
    """API Endpoint to stop MATLAB

    Args:
        req (HTTPRequest): HTTPRequest Object.

    Returns:
        JSONResponse: JSONResponse object containing updated information on the state of MATLAB among other information.
    """
    state = req.app["state"]

    await state.stop_matlab()

    return await create_status_response(req.app)


async def set_licensing_info(req):
    """API Endpoint to set licensing information on the server side.

    Args:
        req (HTTPRequest): HTTPRequest Object

    Raises:
        Exception: If Licensing type is neither MHLM nor NLM.
        web.HTTPBadRequest: If any error with Licensing.

    Returns:
        JSONResponse: JSONResponse object containing updated information on the state of MATLAB among other information.
    """
    state = req.app["state"]

    data = await req.json()
    lic_type = data.get("type")

    try:
        if lic_type == "NLM":
            await state.set_licensing_nlm(data.get("connectionString"))

        elif lic_type == "MHLM":
            await state.set_licensing_mhlm(
                data.get("token"), data.get("emailAddress"), data.get("sourceId")
            )
        else:
            raise Exception('License type must be "NLM" or "MHLM"!')
    except Exception as e:
        raise web.HTTPBadRequest(text="Error with licensing!")

    # Start MATLAB if licensing is complete
    if state.is_licensed() is True and not isinstance(state.error, LicensingError):

        # Start MATLAB
        await state.start_matlab(restart_matlab=True)

    return await create_status_response(req.app)


async def licensing_info_delete(req):
    """API Endpoint to stop MATLAB and remove licensing details.

    Args:
        req (HTTPRequest): HTTPRequest Object
    Returns:
        JSONResponse: JSONResponse object containing updated information on the state of MATLAB among other information.
    """
    state = req.app["state"]

    # Removing license information implies terminating MATLAB
    await state.stop_matlab()

    # Unset licensing information
    state.unset_licensing()

    # Persist licensing information
    state.persist_licensing()

    return await create_status_response(req.app)


async def termination_integration_delete(req):
    """API Endpoint to terminate the Integration and shutdown the server.

    Args:
        req (HTTPRequest): HTTPRequest Object
    """
    state = req.app["state"]

    # Send response manually because this has to happen before the application exits
    res = await create_status_response(req.app, "../")
    await res.prepare(req)
    await res.write_eof()

    # End termination with 0 exit code to indicate intentional termination
    await req.app.shutdown()
    await req.app.cleanup()
    """When testing with pytest, its not possible to catch sys.exit(0) using the construct 
    'with pytest.raises()', there by causing the test : test_termination_integration_delete() 
    to fail. Inorder to avoid this, adding the below if condition to check to skip sys.exit(0) when testing
    """
    if not mwi_env.is_testing_mode_enabled():
        sys.exit(0)


async def root_redirect(request):
    """API Endpoint to return the root index.html file.

    Args:
        request (HTTPRequest): HTTPRequest Object

    Returns:
        HTTPResponse: HTTPResponse Object containing the index.html file.
    """
    base_url = request.app["settings"]["base_url"]
    return aiohttp.web.HTTPFound(f"{base_url}/index.html")


async def static_get(req):
    """Returns HTTP Response objects for the static files

    Args:
        req (HTTPRequest): HTTPRequest object

    Returns:
        HTTPResponse: HTTPResponse object containing the static file.
    """
    details = req.app["static_route_table"][req.path]
    return web.Response(
        headers=details["headers"],
        status=200,
        body=pkgutil.get_data(details["mod"], details["name"]),
    )


def make_static_route_table(app):
    """Traverses the built static files in ./gui folder and adds to a Dict.
    Key for the Dict is the complete path to a file
    Value for the Dict is the header and other information.

    Args:
        app (aiohttp server): The aiohttp server.

    Returns:
        Dict: Containing information about the static files and header information.
    """
    from pkg_resources import resource_isdir, resource_listdir

    from matlab_proxy import gui
    from matlab_proxy.gui import static
    from matlab_proxy.gui.static import css, js, media

    base_url = app["settings"]["base_url"]

    table = {}

    for (mod, parent) in [
        (gui.__name__, ""),
        (gui.static.__name__, "/static"),
        (gui.static.css.__name__, "/static/css"),
        (gui.static.js.__name__, "/static/js"),
        (gui.static.media.__name__, "/static/media"),
    ]:
        for name in resource_listdir(mod, ""):
            if not resource_isdir(mod, name):
                if name != "__init__.py":

                    # Special case for manifest.json
                    if "manifest.json" in name:
                        content_type = "application/manifest+json"
                    else:
                        content_type = mimetypes.guess_type(name)[0]

                    headers = {"content-type": content_type}
                    headers.update(app["settings"]["mwi_custom_http_headers"])

                    table[f"{base_url}{parent}/{name}"] = {
                        "mod": mod,
                        "name": name,
                        "headers": headers,
                    }

    return table


async def matlab_view(req):
    """API Endpoint which proxies requests to the MATLAB Embedded Connector

    Args:
        req (HTTPRequest): HTTPRequest Object

    Raises:
        ValueError: When unable to handle WebSocket Request.
        web.HTTPNotFound: When a non-existing file is requested.

    Returns:
        WebSocketResponse or HTTPResponse: based on the Request type.
    """
    reqH = req.headers.copy()

    state = req.app["state"]
    matlab_port = state.matlab_port
    matlab_protocol = req.app["settings"]["matlab_protocol"]
    mwapikey = req.app["settings"]["mwapikey"]
    matlab_base_url = f"{matlab_protocol}://localhost:{matlab_port}"

    # WebSocket
    if (
        reqH.get("connection") == "Upgrade"
        and reqH.get("upgrade") == "websocket"
        and req.method == "GET"
    ):
        ws_server = web.WebSocketResponse()
        await ws_server.prepare(req)

        async with aiohttp.ClientSession(
            cookies=req.cookies, connector=aiohttp.TCPConnector(verify_ssl=False)
        ) as client_session:

            async with client_session.ws_connect(
                matlab_base_url + req.path_qs,
            ) as ws_client:

                async def wsforward(ws_from, ws_to):
                    async for msg in ws_from:
                        mt = msg.type
                        md = msg.data
                        if mt == aiohttp.WSMsgType.TEXT:
                            await ws_to.send_str(md)
                        elif mt == aiohttp.WSMsgType.BINARY:
                            await ws_to.send_bytes(md)
                        elif mt == aiohttp.WSMsgType.PING:
                            await ws_to.ping()
                        elif mt == aiohttp.WSMsgType.PONG:
                            await ws_to.pong()
                        elif ws_to.closed:
                            await ws_to.close(code=ws_to.close_code, message=msg.extra)
                        else:
                            raise ValueError(f"Unexpected message type: {msg}")

                await asyncio.wait(
                    [wsforward(ws_server, ws_client), wsforward(ws_client, ws_server)],
                    return_when=asyncio.FIRST_COMPLETED,
                )

                return ws_server

    # Standard HTTP Request
    else:
        # Proxy, injecting request header
        async with aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(verify_ssl=False),
        ) as client_session:
            try:
                req_body = await transform_body(req)
                # Set content length in case of modification
                reqH["Content-Length"] = str(len(req_body))
                reqH["x-forwarded-proto"] = "http"

                async with client_session.request(
                    req.method,
                    f"{matlab_base_url}{req.rel_url}",
                    headers={**reqH, **{"mwapikey": mwapikey}},
                    allow_redirects=False,
                    data=req_body,
                ) as res:

                    headers = res.headers.copy()
                    body = await res.read()
                    headers.update(req.app["settings"]["mwi_custom_http_headers"])

                    return web.Response(headers=headers, status=res.status, body=body)
            except Exception:
                raise web.HTTPNotFound()


async def transform_body(req):
    """Transform HTTP POST requests as required by the MATLAB JavaScript Desktop.

    Args:
        req (HTTPRequest): HTTPRequest Object.

    Returns:
        String: String containing JSON object representing the body of a HTTPResponse.
    """

    body = await req.read()

    # Only attempt to rewrite requests known to need rewriting
    if req.method == "POST" and req.rel_url.path.endswith("messageservice/json/secure"):
        data = json.loads(body)
        # Change messages.ClientType.properties.TYPE if necessary
        try:
            replace = False
            for client_type in data["messages"]["ClientType"]:
                if client_type["properties"]["TYPE"] == "jsd":
                    client_type["properties"]["TYPE"] = "jsd_rmt_tmw"
                    replace = True

            if replace is True:
                body = json.dumps(data)
        except KeyError:
            pass

    return body


async def license_init(app):
    """Initializes licensing for the app and is one of the starter tasks when
    the server is started.

    Args:
        app (aiohttp server): The aiohttp server.
    """
    state = app["state"]

    try:
        await state.init_licensing()
    except asyncio.CancelledError:
        pass


async def matlab_starter(app):
    """Upon app startup, start MATLAB if able to do so.

    Args:
        app (aiohttp server): The aiohttp server.
    """
    state = app["state"]

    try:
        if state.is_licensed() and await state.get_matlab_state() == "down":
            await state.start_matlab()
    except asyncio.CancelledError:
        # Ensure MATLAB is terminated
        await state.stop_matlab()


async def start_background_tasks(app):
    """Runs startup tasks asynchronously.
    Initiates licensing and starts matlab asynchronously.

    Args:
        app (aiohttp_server): Instance of aiohttp server
    """
    await license_init(app)
    await matlab_starter(app)


async def cleanup_background_tasks(app):
    """Runs cleanup tasks asynchronously.
    Stops any running tasks and stops matlab asynchronously.

    Args:
        app (aiohttp_server): Instance of aiohttp server
    """
    # First stop matlab
    state = app["state"]
    await state.stop_matlab()

    # Stop any running async tasks
    logger = mwi.logger.get()
    tasks = state.tasks
    for task_name, task in tasks.items():
        if not task.cancelled():
            logger.debug(f"Cancelling MWI task: {task_name} : {task} ")
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass


def configure_and_start(app):
    """Configure the site for the app and update app with appropriate values

    Args:
        app (aiohttp.web.Application): A aiohttp web server.

    Returns:
        aiohttp.web.Application: Updated web server.
    """
    loop = util.get_event_loop()

    web_logger = None if not mwi_env.is_web_logging_enabled() else logger

    # Setup runner
    runner = web.AppRunner(app, logger=web_logger, access_log=web_logger)
    loop.run_until_complete(runner.setup())

    # Prepare site to start, then set port of the app.
    site = util.prepare_site(app, runner)

    # This would be required when MWI_APP_PORT env variable is not set and the site starts on a random port.
    app["settings"]["app_port"] = site._port

    # Update the site origin in settings.
    # The origin will be used for communicating with the Embedded connector.
    app["settings"]["mwi_server_url"] = site.name + app["settings"]["base_url"]

    loop.run_until_complete(site.start())

    logger.debug("Starting MATLAB proxy app")
    logger.debug(
        f' with base_url: {app["settings"]["base_url"]} and app_port:{app["settings"]["app_port"]}.'
    )

    logger.info(
        util.prettify(
            boundary_filler="=",
            text_arr=[f"MATLAB can be accessed at:", app["settings"]["mwi_server_url"]],
        )
    )

    # Startup tasks are being done here as app.on_startup leads
    # to a race condition for mwi_server_url information which is
    # extracted from the site info.
    loop.run_until_complete(start_background_tasks(app))
    return app


# config has a default initializer because it needs to be callable without inputs from ADEV servers
def create_app(config_name=matlab_proxy.get_default_config_name()):
    """Creates the web server and adds the routes,settings and env_config to the server.

    Returns:
        aiohttp server: An aiohttp server with routes, settings and env_config.
    """
    app = web.Application()

    # Get application settings
    app["settings"] = settings.get(
        config_name, dev=mwi_env.is_development_mode_enabled()
    )

    # Initialise application state
    app["state"] = AppState(app["settings"])

    # In development mode, the node development server proxies requests to this
    # development server instead of serving the static files directly
    if not mwi_env.is_development_mode_enabled():
        app["static_route_table"] = make_static_route_table(app)
        for key in app["static_route_table"].keys():
            app.router.add_route("GET", key, static_get)

    base_url = app["settings"]["base_url"]
    app.router.add_route("GET", f"{base_url}/get_status", get_status)
    app.router.add_route("GET", f"{base_url}/get_env_config", get_env_config)
    app.router.add_route("PUT", f"{base_url}/start_matlab", start_matlab)
    app.router.add_route("DELETE", f"{base_url}/stop_matlab", stop_matlab)
    app.router.add_route("PUT", f"{base_url}/set_licensing_info", set_licensing_info)
    app.router.add_route(
        "DELETE", f"{base_url}/set_licensing_info", licensing_info_delete
    )
    app.router.add_route(
        "DELETE", f"{base_url}/terminate_integration", termination_integration_delete
    )
    app.router.add_route("*", f"{base_url}/", root_redirect)
    app.router.add_route("*", f"{base_url}", root_redirect)
    app.router.add_route("*", f"{base_url}/{{proxyPath:.*}}", matlab_view)

    app.on_cleanup.append(cleanup_background_tasks)

    return app


def main():
    """Starting point of the integration. Creates the web app and runs indefinitely."""

    # The integration needs to be called with --config flag.
    # Parse the passed cli arguments.
    desired_configuration_name = util.parse_cli_args()["config"]

    # Create, configure and start the app.
    app = create_app(config_name=desired_configuration_name)
    app = configure_and_start(app)

    loop = util.get_event_loop()
    # Add signal handlers for the current python process
    loop = util.add_signal_handlers(loop)
    loop.run_forever()

    async def shutdown():
        """Shuts down the app in the event of a signal interrupt."""
        logger.info("Shutting down MATLAB proxy-app")
        for task in asyncio.Task.all_tasks():
            logger.debug(f"calling cancel on all_tasks: {task}")
            task.cancel()
        await app.shutdown()
        await app.cleanup()

        asyncio.ensure_future(exit())

    try:
        loop.run_until_complete(shutdown())
    except:
        pass

    logger.info("Finished shutting down. Thank you for using the MATLAB proxy.")
    loop.close()
    sys.exit(0)



# Copyright 2020-2021 The MathWorks, Inc.
import matlab_proxy

# Configure matlab_proxy
config = {
    # Link the documentation url here. This will show up on the website UI
    # where users can create issue's or make enhancement requests
    "doc_url": "https://github.com/mathworks/matlab-proxy/",
    # Use a single word for extension_name
    # It will be used as a flag when launching the integration.
    # NOTE: This name must be used when setting the entrypoint for matlab_proxy in setup.py
    # Use '-' or '_' seperated values if more than 1 word is used.
    # Ex: Hello-World, Alice_Bob.
    "extension_name": matlab_proxy.get_default_config_name(),
    # This value will be used in various places on the website UI.
    # Ensure that this is not more than 3 words.
    "extension_name_short_description": "MATLAB Desktop",
}



# Copyright 2020-2021 The MathWorks, Inc.
"""This file contains validators for various runtime artefacts.
A validator is defined as a function which verifies the input and 
returns it unchanged if validation passes. 
Returning inputs allows validators to be used inline with the input.

Example: 
Original code: if( input ):
With validator: if (valid(input)):

Exceptions are thrown to signal failure.
"""
import errno
import os
import socket
import sys

import matlab_proxy
import pkg_resources

from . import environment_variables as mwi_env
from . import logger as mwi_logger

logger = mwi_logger.get()


def validate_mlm_license_file(nlm_conn_str):
    """Validates and returns input if it passes validation.
    Throws exception when validation fails.
    The connection string should be in the form of port@hostname
    OR path to valid license file

    Args:
        nlm_conn_str (String): Contains the Network license manager connection string.

    Raises:
        NetworkLicensingError: A custom exception.

    Returns:
        String: Returns the same argument passed to this function if its valid.
    """
    import re

    from .exceptions import NetworkLicensingError

    if nlm_conn_str is None:
        return None

    # TODO: The JS validation of this setting does not allow file path locations
    # The JS validation occurs before reaching the set_licensing_info endpoint.

    # Regular expression to match port@hostname,
    # where port is any number and hostname is alphanumeric
    # regex = Start of Line, Any number of 0-9 digits , @, any number of nonwhite space characters with "- _ ." allowed
    # "^[0-9]+[@](\w|\_|\-|\.)+$"
    # Server triad is of the form : port@host1,port@host2,port@host3
    regex = "(^[0-9]+[@](\w|\_|\-|\.)+$)|(^[0-9]+[@](\w|\_|\-|\.)+),([0-9]+[@](\w|\_|\-|\.)+),([0-9]+[@](\w|\_|\-|\.)+$)"
    if not re.search(regex, nlm_conn_str):
        logger.debug("NLM info is not in the form of port@hostname")
        if not os.path.isfile(nlm_conn_str):
            logger.debug("NLM info is not a valid path to a license file")
            error_message = (
                f"MLM_LICENSE_FILE validation failed for {nlm_conn_str}. "
                f"If set, the MLM_LICENSE_FILE environment variable must be a string which is either of the form port@hostname"
                f" OR path to a valid license file."
            )
            logger.error(error_message)
            raise NetworkLicensingError(error_message)
        else:
            logger.info(
                f"MLM_LICENSE_FILE with value: {nlm_conn_str} is a path to a file. MATLAB will attempt to use it."
            )
    else:
        logger.info(
            f"MLM_LICENSE_FILE with value: {nlm_conn_str} is a license server, MATLAB will attempt to connect to it."
        )

    # Validation passed
    return nlm_conn_str


def validate_app_port_is_free(port):
    """Validates and returns port if its free else will error out and exit.

    Args:
        port (str|int): Port number either as a string or an integer.

    Raises:
        e: socket.error

    Returns:
        Boolean: True if provided port is occupied else False.
    """
    # If port is None, at launch, site will use a randomnly allocated port.
    if port is None:
        logger.debug(
            f"Environment variable {mwi_env.get_env_name_app_port()} was not set. Will use a random port at launch."
        )
        return port

    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind(("", int(port)))
        s.close()

        # Was able to allocate port. Validation passed.
        return port
    except socket.error as e:
        if e.errno == errno.EADDRINUSE:
            logger.error(
                f"The port {port} is not available. Please set another value for the environment variable {mwi_env.get_env_name_app_port()}"
            )
            sys.exit(1)
        else:
            raise e


def validate_base_url(base_url):
    """Validates base_url set in the env variable MWI_BASE_URL.

    If MWI_BASE_URL is empty, will use base_url="/"
    If MWI_BASE_URL doesnt have a prefix '/' will error out.
    If MWI_BASE_URL has a suffix '/', will remove it.

    Args:
        base_url (str | None): The base_url at which the integration will be launched.

    Returns:
        [str]: Validated base_url
    """
    validated_base_url = ""
    if base_url == "":
        validated_base_url = ""

    else:
        if not base_url.startswith("/"):
            logger.error(
                f'The value of environment variable {mwi_env.get_env_name_base_url()} must start with "/" '
            )
            sys.exit(1)

        validated_base_url = base_url[:-1] if base_url.endswith("/") else base_url

    return validated_base_url


def validate_env_config(config):
    """Validates config passed with available "matlab_proxy_configs" entry point in the same
    python environment. Computes DDUX value for MATLAB use.

    Args:
        config (str): Name of the configuration to use.

    Returns:
        Dict: Containing data specific to the environment in which MATLAB proxy is being used in.
    """
    available_configs = __get_configs()
    config = config.lower()

    # Check if supplied config is present in the available configs
    if config in available_configs:
        # Check if all keys are present in the supplied config
        default_config_keys = available_configs[
            matlab_proxy.get_default_config_name()
        ].keys()
        env_config = available_configs[config]

        for key in default_config_keys:
            if not key in env_config:
                logger.error(f"{key} missing in the provided {config} configuration")
                sys.exit(1)

        logger.debug(f"Successfully validated provided {config} configuration")
        return env_config

    else:
        logger.error(
            f"{config} is not a valid config. Available configs are : {list(available_configs.keys())}"
        )
        sys.exit(1)


def __get_configs():
    """Iterates over the 'entry_points' of the installed packages in the current python
    environment and loads the 'matlab_proxy_configs' entry point values into the 'configs' Dict.

    Returns:
        Dict: Contains all the values present in 'matlab_web_desktop_configs' entry_point from all the packages
        installed in the current environment.
    """
    configs = {}
    for entry_point in pkg_resources.iter_entry_points(
        matlab_proxy.get_entrypoint_name()
    ):
        configs[entry_point.name.lower()] = entry_point.load()

    return configs


def validate_ssl_cert_file(a_ssl_cert_file):
    """Ensures that its a valid readable file"""

    # Empty strings are valid inputs
    if a_ssl_cert_file:
        # String is not empty, check to see if the file exists
        if not os.path.isfile(a_ssl_cert_file):
            logger.error(f"MWI_SSL_CERT_FILE is not a valid file: {a_ssl_cert_file}")
            sys.exit(1)

    # string is either empty, or is a valid file on disk
    return a_ssl_cert_file


def validate_ssl_key_and_cert_file(a_ssl_key_file, a_ssl_cert_file):
    """Ensures that its a valid readable file"""

    if a_ssl_cert_file is None and a_ssl_key_file is None:
        # Both values are None, this is acceptable.
        return a_ssl_key_file, a_ssl_cert_file

    # Implies atleast one value is not None.

    # Cert file is either empty or valid file.
    cert_file = validate_ssl_cert_file(a_ssl_cert_file=a_ssl_cert_file)

    if cert_file is None and a_ssl_key_file is not None:
        logger.error(f"MWI_SSL_CERT_FILE must be provided to use the MWI_SSL_KEY_FILE")
        sys.exit(1)

    if a_ssl_key_file is None and cert_file is not None:
        logger.info(
            f"MWI_SSL_KEY_FILE is not provided, ensure that your MWI_SSL_CERT_FILE : '{cert_file}' contains a private key"
        )

    if a_ssl_key_file:
        if not os.path.isfile(a_ssl_key_file):
            logger.error(f"MWI_SSL_KEY_FILE is not a valid file: {a_ssl_key_file}")
            sys.exit(1)

    logger.info(
        f"SSL Keys provided were: MWI_SSL_CERT_FILE: {a_ssl_cert_file} & MWI_SSL_KEY_FILE: {a_ssl_key_file}"
    )
    return a_ssl_key_file, a_ssl_cert_file



# Copyright 2021 The MathWorks, Inc.
"""This file lists and exposes the environment variables which are used by the integration."""

import os


def get_env_name_network_license_manager():
    """Specifies the path to valid license file or address of a network license server"""
    return "MLM_LICENSE_FILE"


def get_env_name_mhlm_context():
    """Specifies the context from which MHLM was initiated. Used by DDUX in MATLAB."""
    return "MHLM_CONTEXT"


def get_env_name_logging_level():
    """Specifies the logging level used by app's loggers"""
    return "MWI_LOG_LEVEL"


def get_env_name_web_logging_enabled():
    """Enable the logging of asyncio web traffic by setting to true"""
    return "MWI_WEB_LOGGING_ENABLED"


def get_env_name_log_file():
    """Specifies a file into which logging content is directed"""
    return "MWI_LOG_FILE"


def get_env_name_base_url():
    """Specifies the base url on which the website should run.
    Eg: www.127.0.0.1:8888/base_url/index.html

    Note: The website runs on a URL of the form:
        www.<SERVER ADDRESS>:<PORT NUMBER>/<BASE_URL>/index.html

    Note: If you are updating this value, remember to update the startup.m file
            that is used to notify the connector of the base url.
    """
    return "MWI_BASE_URL"


def get_env_name_app_port():
    """Specifies the port on which the website is running on the server.
    Eg: www.127.0.0.1:PORT/index.html

    Note: The website runs on a URL of the form:
        www.<SERVER ADDRESS>:<PORT NUMBER>/<BASE_URL>/index.html
    """
    return "MWI_APP_PORT"


def get_env_name_custom_http_headers():
    """Specifies HTTP headers as JSON content, to be injected into responses sent to the browser"""
    return "MWI_CUSTOM_HTTP_HEADERS"


def get_env_name_app_host():
    """Specifies the host on which the TCP site (aiohttp server) is being run."""
    return "MWI_APP_HOST"


def get_env_name_testing():
    """Set to true when we are running tests in development mode."""
    return "MWI_TEST"


def get_env_name_development():
    """Set to true when we are in development mode."""
    return "MWI_DEV"


def get_env_name_matlab_tempdir():
    """The environment variables used to control the temp directory used by MATLAB on POSIX systems"""
    # Order matters, MATLAB checks TMPDIR first and then TMP
    return ["TMPDIR", "TMP"]


def is_development_mode_enabled():
    """Returns true if the app is in development mode."""
    return os.environ.get(get_env_name_development(), "false").lower() == "true"


def is_testing_mode_enabled():
    """Returns true if the app is in testing mode."""
    return (
        is_development_mode_enabled()
        and os.environ.get(get_env_name_testing(), "false").lower() == "true"
    )


def is_web_logging_enabled():
    """Returns true if the web logging is required to be enabled"""
    return os.environ.get(get_env_name_web_logging_enabled(), "false").lower() == "true"


def get_env_name_ssl_cert_file():
    """Specifies the certificate to be used by webserver."""
    return "MWI_SSL_CERT_FILE"


def get_env_name_ssl_key_file():
    """Specifies the key used by webserver to sign the ssl certificate."""
    return "MWI_SSL_KEY_FILE"
